# pylint: disable=import-error, too-few-public-methods, unused-import, too-many-arguments, line-too-long
"""
    CUSTOM WRITER CLASSES
        - Class which manages writer tasks like
        auth, write metadata, write file, create dir structure
"""
import re
import json
from typing import Union, List, Any, Dict, Tuple, Optional
import boto3
from sdc_dp_helpers.base_writers import BaseWriter


class GoogleKnowledgeGraphWriter(BaseWriter):
    """GoogleKnowledgeGraphReader Class"""

    def __init__(
        self,
        bucket: str,
        folder_path: str,
        destination: str,
        configs: Optional[Dict[str, Any]] = None,
        clear_destination: bool = False,
    ):
        self.success: List[bool] = []
        self.current_destination: Union[str, None] = None
        super().__init__(
            bucket=bucket,
            folder_path=folder_path,
            destination=destination,
            configs=configs,
            clear_destination=clear_destination,
        )

    def verify_data(
        self, payload: Dict[Any, Any]
    ) -> Tuple[str, Union[List[Dict[Any, Any]], Dict[Any, Any], Any]]:
        """verifies data before writting to destination also designs the destination file path

        Args:
            payload (Dict[Any, Any]):

        Raises:
            KeyError: if we do not find the exact keys we expect from the payload
            TypeError: if provided data object is not a list

        Returns:
            Tuple[str, Union[List[Dict[Any, Any]], Dict[Any, Any], Any]]: full destination path and
                                                                          data to be written
        """
        # confirm the payload keys are matching accurately with what is expected
        if not {"data", "date"}.issubset(set(payload.keys())):
            raise KeyError("Invalid payload ensure you have date and data as keys")

        if not re.search(r"\d{4}\-\d{2}\-\d{2}", payload["date"]):
            raise ValueError(
                "Invalid date format for partitioning expected: 'YYYY-mm-dd'"
            )
        if not isinstance(payload["data"], list):
            raise TypeError("Invalid data passed: expected List[Dict, Dict]")
        _date = payload["date"].replace("-", "")
        _data = payload["data"]
        write_path = f"{self.folder_path}/{_date}"
        return write_path, _data

    def write_stats_to_s3(
        self,
        stats: str,
        date: str,
        vertical: str,
        target_table_name: str = "gkg_api_statistics"
    ) -> None:
        """
        Write statistics about the number of api calls on a particular date, network and endpoint.
        """
        boto3_session = boto3.Session()
        s3_resource = boto3_session.resource("s3")
        write_path: str = f"{target_table_name}/{vertical}/{date}.json"
        if stats:
            print(
                f"Writing data to s3://{self.bucket}/{write_path} partitioned by networks and date."
            )
            s3_resource.Object(self.bucket, write_path).put(
                Body=json.dumps(stats))



import re


def multiple_regex_replace(text: str, replace_map: dict) -> str:
    """Replace multiple Regex Patterns in one go
    @param: replace_map - dict of key (what to replace) value (replacements)
                            i.e {".": "_dot_", "+": "_plus_", " ":"_space_"}
    @param: text - string to perform the regex substititution on.
    """
    # Create a regular expression  from the dictionaryionary keys
    # regex = re.compile("(%s)" % "|".join(map(re.escape, replace_map.keys())))
    regex = re.compile(f"({'|'.join(map(re.escape, replace_map.keys()))})")

    # For each match, look-up corresponding value in dictionaryionary
    return regex.sub(lambda mo: replace_map[mo.string[mo.start() : mo.end()]], text)



"""CUSTOM INTERFACES FOR ZOHO RECRUIT ENDPOINTS"""

# pylint: disable=broad-except,too-few-public-methods,too-many-arguments,arguments-differ, too-many-locals,abstract-method,unnecessary-dict-index-lookup,import-error
import sys
import os

from datetime import datetime, timedelta
from abc import ABC, abstractmethod
from typing import Union, Any, List, Dict, Generator, Tuple

from io import BytesIO
from io import TextIOWrapper
from csv import DictReader
from zipfile import ZipFile

import requests

from requests.exceptions import HTTPError
from sdc_dp_helpers.api_utilities.date_managers import date_string_handler
from sdc_dp_helpers.api_utilities.retry_managers import retry_handler, request_handler


class AuthenticationError(Exception):
    """Raised when our access token is invalid message INVALID_TOKEN"""


class APILimitError(Exception):
    """Error raised for exceeding API limits"""


class Authenticator:
    """Authenticator Class for Zoho Recruit"""

    def __init__(self):
        self.next_refresh: datetime = datetime.now()
        self.access_token: str = ""

    def get_access_token(
        self, sessions: requests.Session, secrets: Dict[Any, Any]
    ) -> str:
        """Authentication method

        Args:
            sessions (requests.Session): requests.Sessions
            secrets (Dict[Any, Any]): dictionary having secrets

        Returns:
            str : the access_token for zoho
        """
        response = {
            "access_token": None,
            "api_domain": "https://www.zohoapis.com",
            "token_type": "Bearer",
            "expires_in": 0,
        }
        try:
            response = sessions.post(
                url="https://accounts.zoho.com/oauth/v2/token?",
                params={
                    "refresh_token": secrets["refresh_token"],
                    "client_id": secrets["client_id"],
                    "client_secret": secrets["client_secret"],
                    "grant_type": "refresh_token",
                },
                timeout=61,
            ).json()
        except Exception as exc:
            print(f"Error on refreshing oauth_token {exc}")
        seconds_to_expiry = int(response["expires_in"]) - 100
        self.next_refresh = datetime.now() + timedelta(seconds=seconds_to_expiry)

        print(f"new oauth_token {response['access_token']}")
        print(f"next oauth_token_refresh {self.next_refresh}")
        self.access_token = response["access_token"]
        return response["access_token"]

    def get_refresh_token(
        self, sessions: requests.Session, secrets: Dict[str, str]
    ) -> Dict[str, Any]:
        """Gets refresh token and access toekn used by the api calls see url:
        https://www.zoho.com/recruit/developer-guide/apiv2/access-refresh.html

        Args:
            configs (Dict[str, str]): client_id, client_secret, code, grant_type, redirect_uri

        Returns:
            Dict[str, Any]: response with refresh_token and access_token
        """
        results = {
            "access_token": None,
            "refresh_token": None,
            "api_domain": "https://www.zohoapis.com",
            "token_type": "Bearer",
            "expires_in": 3600,
        }
        response = sessions.post(
            url=secrets.get("url", "https://accounts.zoho.com/oauth/v2/token"),
            params={
                "grant_type": "authorization_code",
                "client_id": secrets["client_id"],
                "client_secret": secrets["cleint_secrets"],
                "code": secrets["code"],
                "redirect_uri": "https://accounts.zoho.com/developerconsole",
            },
            timeout=60,
        )
        status_code = response.status_code
        if status_code != 200:
            print(
                f"error getting data from: {response.url} \nerror_details: {response.json()}\n"
            )
        if status_code == 200:
            results = response.json()
        return results

    def authenticate(self, **kwargs):
        """Authenticate Zoho Access"""
        if self.next_refresh <= datetime.now():
            self.get_access_token(**kwargs)
        return self.access_token


class RequestHandler(ABC):
    """Interface for API Call Method"""

    API_LIMIT_EXCEEDED_MESSAGE = "Exceeded API limits going to sleep"

    def __init__(self, session: requests.Session):
        self.session = session

    @abstractmethod
    def make_api_call(self, **kwargs):
        """Make API Call"""
        raise NotImplementedError


class CreateBulkJob(RequestHandler):
    """Creates Bulk Job"""

    @retry_handler(exceptions=APILimitError, total_tries=5, initial_wait=180)
    @request_handler(
        wait=int(os.environ.get("API_WAIT_TIME", 30)),
        backoff_factor=0.01,
        backoff_method="random",
    )
    def make_api_call(
        self, access_token: str, module: str, page: int, criteria: dict
    ) -> Tuple[Union[str, None], Union[str, None], Dict[Any, Any]]:
        """Create Bulk Job"""
        status, job_id, details = None, None, {"id": None}
        query = {"module": module, "page": page}
        query.update(criteria)
        print(query)
        response = self.session.post(
            url="https://recruit.zoho.com/recruit/bulk/v2/read",
            json={
                "query": query,
            },
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Zoho-oauthtoken {access_token}",
                # "If-Modified-Since": modified_since,
            },
            timeout=61,
        )
        status_code = response.status_code
        if status_code == 429:
            raise APILimitError(self.API_LIMIT_EXCEEDED_MESSAGE)

        if status_code not in [200, 201]:
            print(
                f"error getting data from: {response.url} \nerror_details: {response.json()}\n"
            )

        print(
            f"{query['module']} job creation attempt done with status_code: {status_code}"
        )
        if response.status_code in [200, 201]:
            json_data = response.json().get("data", [])[0]
            status = json_data["status"]
            job_id, details = json_data["details"]["id"], json_data["details"]

        return status, job_id, details


class CheckBulkJobStatus(RequestHandler):
    """Request Call to Check Bulk Job Status"""

    @retry_handler(exceptions=APILimitError, total_tries=5, initial_wait=180)
    @request_handler(
        wait=int(os.environ.get("API_WAIT_TIME", 30)),
        backoff_factor=0.01,
        backoff_method="random",
    )
    def make_api_call(
        self, access_token: str, job_id: str
    ) -> Tuple[Union[str, None], bool, Dict[Any, Any]]:
        """Check status of bulk job"""

        results = {"data": [{"id": job_id, "state": None, "result": {}}]}
        state, more_records = None, False
        response = self.session.get(
            url=f"https://recruit.zoho.com/recruit/bulk/v2/read/{job_id}",
            headers={
                "Authorization": f"Zoho-oauthtoken {access_token}",
            },
            timeout=61,
        )
        status_code = response.status_code
        if status_code == 429:
            raise APILimitError(self.API_LIMIT_EXCEEDED_MESSAGE)

        if status_code != 200:
            print(
                f"error getting data from: {response.url} \nerror_details: {response.json()}\n"
            )

        if status_code == 200:
            json_data = response.json()["data"][0]
            results = json_data.get("result", {"more_records": False})
            state, more_records = json_data["state"], results["more_records"]
            # page, record_count = results["page"], results["count"]
        return state, more_records, results


class BulkRequestHandler(RequestHandler):
    """Bulk Result Reader"""

    @retry_handler(exceptions=APILimitError, total_tries=5, initial_wait=180)
    @request_handler(
        wait=int(os.environ.get("API_WAIT_TIME", 30)),
        backoff_factor=0.01,
        backoff_method="random",
    )
    def make_api_call(self, access_token: str, job_id: str):
        """Download Bilk Data"""
        results = None
        response = self.session.get(
            url=f"https://recruit.zoho.com/recruit/bulk/v2/read/{job_id}/result",
            headers={
                "Authorization": f"Zoho-oauthtoken {access_token}",
            },
            timeout=61,
        )
        status_code, results = response.status_code, None
        if status_code == 429:
            raise APILimitError(self.API_LIMIT_EXCEEDED_MESSAGE)

        if status_code not in [200, 201]:
            print(
                f"error getting data from: {response.url}"
                f"status_code: {status_code}\nerror_details: {response.json()}\n"
            )

        if status_code:
            results = response.content
        return results


class AssessmentRequestHandler(RequestHandler):
    """Class for Making API Call"""

    @retry_handler(exceptions=APILimitError, total_tries=5, initial_wait=180)
    @request_handler(
        wait=int(os.environ.get("API_WAIT_TIME", 30)),
        backoff_factor=0.01,
        backoff_method="random",
    )
    def make_api_call(
        self,
        access_token: str,
        url: str,
        page: int,
        per_page: int,
        sort_by: str,
        sort_order: str,
        modified_since: Union[str, None],
        **kwargs,
    ):
        response = self.session.get(
            url=url,
            json={
                "page": page,
                "per_page": per_page,
                "sort_by": sort_by,
                "sort_order": sort_order,
            },
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Zoho-oauthtoken {access_token}",
                "If-Modified-Since": modified_since,
            },
            timeout=61,
        )

        status_code, reason = response.status_code, response.reason
        if status_code == 429:
            raise APILimitError(self.API_LIMIT_EXCEEDED_MESSAGE)

        if status_code != 200:
            print(
                f"error getting data from: {response.url}"
                f"status_code: {status_code}\nerror_details: {response.json()}\n"
            )

        data, more_records = None, False
        if response.status_code == 200:
            results = response.json()
            data = results["Assessments"]
            more_records = False
        return data, more_records, status_code, reason


class RecordsRequestHandler(RequestHandler):
    """Class for Making API Call"""

    @retry_handler(exceptions=APILimitError, total_tries=5, initial_wait=180)
    @request_handler(
        wait=int(os.environ.get("API_WAIT_TIME", 30)),
        backoff_factor=0.01,
        backoff_method="random",
    )
    def make_api_call(
        self,
        access_token: str,
        url: str,
        page: int,
        per_page: int,
        sort_by: str,
        sort_order: str,
        modified_since: Union[str, None],
        **kwargs,
    ):
        response = self.session.get(
            url=url,
            params={
                "page": page,
                "per_page": per_page,
                "sort_by": sort_by,
                "sort_order": sort_order,
            },
            headers={
                "Authorization": f"Zoho-oauthtoken {access_token}",
                "If-Modified-Since": modified_since,
            },
            timeout=61,
        )

        status_code, reason = response.status_code, response.reason
        if status_code == 429:
            raise APILimitError(self.API_LIMIT_EXCEEDED_MESSAGE)

        if status_code != 200:
            print(
                f"error getting data from: {response.url}"
                f"status_code: {status_code}\nerror_details: {response.json()}\n"
            )

        data, more_records = None, False
        if response.status_code == 200:
            results = response.json()
            data = results["data"]
            more_records = results.get("info", {"more_records": False}).get(
                "more_records"
            )
        return data, more_records, status_code, reason


class UsersRequestHandler(RequestHandler):
    """Class for Making API Call"""

    @retry_handler(exceptions=APILimitError, total_tries=5, initial_wait=180)
    @request_handler(
        wait=int(os.environ.get("API_WAIT_TIME", 30)),
        backoff_factor=0.01,
        backoff_method="random",
    )
    def make_api_call(
        self,
        access_token: str,
        url: str,
        page: int,
        per_page: int,
        sort_by: str,
        sort_order: str,
        modified_since: Union[str, None],
        **kwargs,
    ):
        response = self.session.get(
            url=url,
            json={
                "type": kwargs.get("users_type"),
                "page": page,
                "per_page": per_page,
                "sort_by": sort_by,
                "sort_order": sort_order,
            },
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Zoho-oauthtoken {access_token}",
                "If-Modified-Since": modified_since,
            },
            timeout=61,
        )

        status_code, reason = response.status_code, response.reason
        if status_code == 429:
            raise APILimitError(self.API_LIMIT_EXCEEDED_MESSAGE)

        if status_code != 200:
            print(
                f"error getting data from: {response.url}"
                f"status_code: {status_code}\nerror_details: {response.json()}\n"
            )

        data, more_records = None, False
        if response.status_code == 200:
            results = response.json()
            data = results["users"]
            more_records = results.get("info", {"more_records": False}).get(
                "more_records"
            )
        return data, more_records, status_code, reason


class APIRequestHandlerFactory:
    """Gets Us the API Call Handler"""

    def get_api_call_handler(self, method: str, session: requests.Session):
        """Gets for use the API Call Handler to Use"""
        api_call_handlers = {
            "assessments": AssessmentRequestHandler,
            "records": RecordsRequestHandler,
            "users": UsersRequestHandler,
            "bulk": BulkRequestHandler,
        }
        return api_call_handlers[method](session=session)


class ZohoRecruitAPI:
    """Sample"""

    authenticator = Authenticator()

    def __init__(
        self,
        session: requests.Session,
        secrets: Dict[Any, Any],
        file_size_limit: int = 104857,  # 5MB
    ):
        self.session = session
        self.secrets = secrets
        self.access_token = self.get_access_token()
        self.file_size_limit = file_size_limit

    def get_access_token(self):
        """Get zoho recruit oauth_token"""
        self.access_token = self.authenticator.authenticate(
            sessions=self.session, secrets=self.secrets
        )
        return self.access_token

    def get_end_point_handler(self, end_point: str):
        """Gets the API Handler to use in getting data"""
        endpoints = {
            "bulk": BulkEndpointHandler,
            "records": RecordsEndpointHandler,
            "users": UsersEndpointHandler,
        }
        if end_point not in endpoints:
            raise KeyError(f"provided endpoint not yet implemented: {end_point}")

        return endpoints[end_point](session=self.session, secrets=self.secrets)


class RestAPICaller(ZohoRecruitAPI):
    """Class to Call the RestAPI Endpoints"""

    def api_call(self, api_call_handler: RequestHandler, **kwargs):
        """Makes API Call"""
        results = None
        status_code = None
        reason = None
        more_records = True

        try:
            response = api_call_handler.make_api_call(
                access_token=self.access_token,
                url=kwargs.get("url"),
                module=kwargs.get("module"),
                page=kwargs.get("page"),
                per_page=kwargs.get("per_page"),
                sort_by=kwargs.get("sort_by"),
                sort_order=kwargs.get("sort_order"),
                modified_since=kwargs.get("modified_since"),
                users_type=kwargs.get("users_type"),
            )
            status_code, reason = response[2], response[3]
            results = response[0]
            more_records = response[1]

        except HTTPError as exc:
            print(
                f"error code: {exc.response.status_code} error message: {exc.response}"
            )
            if exc.response.status_code != 204:
                raise exc
        except requests.exceptions.JSONDecodeError:
            pass
        except Exception as err:
            print(f"failed with status {status_code} reason {reason}\n")
            raise err
        return results, more_records

    def paginate(
        self,
        api_call_handler: RequestHandler,
        module: str,
        configs: Dict[str, Any],
        **kwargs,
    ):  # List[Dict[Any, Any]]:
        """Method to actually make the API call

        Args:
            configs: dict of parameters and headers for the API call

        Returns:
            List[Dict[Any, Any]]: A list of dictionary of each response
        """

        result_array = []
        avg_record_size: int = 0
        current_size: int = 0
        page: int = configs.get("page", 1)
        per_page: int = configs.get("per_page", 200)
        modified_since: Union[None, str] = configs.get("modified_since")
        sort_by = configs.get("sort_by", "Modified_Time")
        sort_order = configs.get("sort_order", "desc")
        users_type = kwargs.get("users_type")
        page_limit: int = configs.get("page_limit", 1000)
        self.file_size_limit = configs.get("file_size_limit", 5242880)
        counter: int = 0
        base_url = "https://recruit.zoho.com/recruit/v2"

        more_records = True
        url = f"{base_url}/{module}"
        while more_records:
            print(
                f"getting module {module} page {page} "
                f"more_records {more_records} current_size {current_size}"
            )

            self.access_token = self.get_access_token()

            response = self.api_call(
                api_call_handler,
                url=url,
                module=module,
                page=page,
                per_page=per_page,
                sort_by=sort_by,
                sort_order=sort_order,
                modified_since=modified_since,
                users_type=users_type,
            )

            results, more_records = response[0], response[1]
            if results is None:
                yield {"start_page": page, "data": result_array}
                break

            if avg_record_size == 0:
                avg_record_size = sys.getsizeof(results[0])

            result_array = result_array + results
            current_size += avg_record_size
            if counter > page_limit:  # exit if we go past page_limit
                print(f"\nExiting, we have reached page_limit {page_limit}\n")
                yield {"start_page": page, "data": result_array}
                break
            counter += 1

            if (current_size >= self.file_size_limit) or (more_records is False):
                yield {"start_page": page, "data": result_array}
                result_array = []
                current_size = 0
            page += 1


class RecordsEndpointHandler(RestAPICaller):
    """Class to Handle Get Records Endpoint"""

    def fetch_data(
        self, configs: dict
    ) -> Generator[Dict[str, Union[str, List[Dict[Any, Any]]]], None, None]:
        """Fetches the dat for records endpoint"""

        api_factory = APIRequestHandlerFactory()
        for module in configs["modules"]:
            method = "records" if module != "Assessments" else "assessments"

            api_call_handler = api_factory.get_api_call_handler(
                method, session=self.session
            )
            results = self.paginate(api_call_handler, module=module, configs=configs)
            for result in results:
                result["module"] = module
                yield result


class UsersEndpointHandler(RestAPICaller):
    """Handler Class for Users Endpoint"""

    def fetch_data(
        self, configs: dict
    ) -> Generator[Dict[str, Union[str, List[Dict[Any, Any]]]], None, None]:
        """Fetches the dat for records endpoint"""

        api_factory = APIRequestHandlerFactory()
        method = "users"
        api_call_handler = api_factory.get_api_call_handler(
            method, session=self.session
        )
        users_types = configs.get("types", ["AllUsers"])
        for users_type in users_types:
            results = self.paginate(
                api_call_handler, module=method, configs=configs, users_type=users_type
            )
            for result in results:
                result["module"] = users_type
                yield result


class BulkAPICaller(ZohoRecruitAPI):
    """Calls the Bulk API"""

    def __init__(
        self,
        secrets: Dict[Any, Any],
        session: requests.Session,
        file_size_limit: int = 5242880,  # 5MB
    ):
        self.jobs: Dict[str, Any] = {}
        self.configs: Dict[Any, Any] = {}
        super().__init__(session, secrets, file_size_limit)
        self.check_status = CheckBulkJobStatus(self.session)
        self.create_job = CreateBulkJob(self.session)

    @staticmethod
    def criteria_datetime_processor(datetime_fields: dict, groups: list) -> list:
        """processor for datetime in criteria"""
        final_groups = []
        for group in groups:
            api_name = group["api_name"]
            if api_name in datetime_fields:
                start = datetime.strftime(
                    date_string_handler(group["value"][0]),
                    datetime_fields.get(api_name, "%Y-%m-%dT00:00:00+00:00"),
                )
                end = datetime.strftime(
                    date_string_handler(group["value"][1]),
                    datetime_fields.get(api_name, "%Y-%m-%dT00:00:00+00:00"),
                )
                group["value"] = [start, end]

            final_groups.append(group)

        start_end = [
            key["value"] for key in final_groups if key["api_name"] == "Modified_Time"
        ]
        if start_end == []:
            raise KeyError("provide `Modified_Time` as a default filter in `group`")
        group = [
            {
                "api_name": "Modified_Time",
                "value": [start_end[0][0], start_end[0][1]],
                "comparator": "between",
            },
        ]

        return final_groups

    @staticmethod
    def build_bulk_query(
        local_configs: dict, criteria_group: list, module: str
    ) -> dict:
        """Method used to build the query to send
        https://www.zoho.com/recruit/developer-guide/apiv2/bulk-read/create-job.html
        """

        group = []
        if module in ["Candidates"]:
            for item in criteria_group:
                if item["api_name"] == "Modified_Time":
                    new_item = {}
                    new_item["api_name"] = "Last_Activity_Time"
                    new_item.update(
                        {key: value for key, value in item.items() if key != "api_name"}
                    )
                    group.append(new_item)
                    continue
                if item["api_name"] == "Created_Time":
                    group.extend([item, item])
                    continue
                group.append(item)
            criteria_group = group

        query = {
            "selectfields": local_configs.get("select_fields", []),
            "criteria": {
                "group_operator": local_configs.get("group_comparator", "and"),
            },
        }
        query["criteria"]["group"] = criteria_group
        if local_configs.get("cvid") is not None:
            query["cvid"] = local_configs.get("cvid")

        return query

    def wait_for_download(self, job_id: str) -> Tuple[bool, int]:
        """Method to keep checking for download status"""

        state: str = ""
        proceed: bool = False
        page: int = 1
        record_count: int = 0
        more_records: bool = False
        total_wait_time, wait_threshold = 0, 600
        while (state != "COMPLETED") or (total_wait_time > wait_threshold):
            self.access_token: str = self.get_access_token()
            state, more_records, results = self.check_status.make_api_call(
                access_token=self.access_token, job_id=job_id
            )
            page, record_count = results.get("page"), results.get("count")

            if state == "COMPLETED":
                proceed = True
                break

        print(
            f"page: {page}, record_count: {record_count}, more_records: {more_records}"
        )
        return proceed, record_count

    @staticmethod
    def parse_zip_file(response):
        """Parse the zip file"""
        # data_array: List[Dict[Any, Any]] = []
        with ZipFile(BytesIO(response)) as zipfile:
            for file in zipfile.filelist:
                with zipfile.open(file) as file_contents:
                    dict_reader = DictReader(
                        TextIOWrapper(file_contents, encoding="utf-8")
                    )
                    yield from dict_reader

    def api_call(self, api_call_handler: RequestHandler, **kwargs):
        """Makes api call"""
        job_id: str = kwargs["job_id"]
        download_status = self.wait_for_download(job_id=job_id)
        if download_status[0] is True and download_status[1] > 0:
            self.access_token = self.get_access_token()
            response = api_call_handler.make_api_call(
                access_token=self.access_token, job_id=job_id
            )
            yield from self.parse_zip_file(response)

    def paginate(self, api_call_handler: RequestHandler, module, configs, criteria):
        """Paginator Function Through the Pages"""

        result_array = []
        current_size: int = 0
        more_records = True
        self.file_size_limit = configs.get("file_size_limit", 5242880)

        page = 1

        while more_records:
            status, job_id, _ = self.create_job.make_api_call(
                self.access_token, module, page, criteria
            )

            print(f"bulk_job status: {status} and job_id : {job_id}")
            state, more_records, _ = self.check_status.make_api_call(
                self.access_token, job_id
            )
            if job_id is not None:
                self.jobs[job_id] = {"state": state, "downloaded": False}

            page += 1

        for job_id, values in self.jobs.items():
            if values["downloaded"] is True:
                continue
            results = self.api_call(api_call_handler, job_id=job_id)
            if not results:
                print(f"No data for job_id {job_id}")
                continue
            counter = 0
            for result in results:
                self.jobs[job_id]["downloaded"] = True
                result_array.append(result)
                current_size += sys.getsizeof(result)
                if current_size >= self.file_size_limit:
                    yield {"start_page": counter, "data": result_array}
                    result_array = []
                    current_size = 0
                counter += 1

            if result_array:
                yield {"start_page": counter, "data": result_array}


class BulkEndpointHandler(BulkAPICaller):
    """Handler for Mkaing the Download Call"""

    def fetch_data(
        self, configs: dict
    ) -> Generator[Dict[str, Union[str, List[Dict[Any, Any]]]], None, None]:
        """Code to fetch the data"""

        method = "bulk"
        criteria = {}
        criteria_group = self.criteria_datetime_processor(
            configs.get("criteria_datetime_fields", {}), configs["group"]
        )
        for module in configs["modules"]:
            criteria = self.build_bulk_query(
                local_configs=configs,
                criteria_group=criteria_group,
                module=module,
            )

            api_factory = APIRequestHandlerFactory()
            api_call_handler = api_factory.get_api_call_handler(
                method, session=self.session
            )
            results = self.paginate(
                api_call_handler, module=module, configs=configs, criteria=criteria
            )
            for result in results:
                result["module"] = module
                yield result



import json
import paramiko
from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.retry_managers import retry_handler
from sdc_dp_helpers.api_utilities.date_managers import date_range, date_string_change_format


class SFTPReader:
    """
    Custom FTP Reader
    """

    def __init__(self, creds_file, config_file):
        self._creds: dict = load_file(creds_file, "yml")

        self.config = load_file(config_file)

        self._connect()

    def _connect(self):
        ssh_client = paramiko.SSHClient()
        ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh_client.connect(hostname=str(self.config.get("hostname")),
                           port=self.config.get("port", 22),
                           username=str(self._creds.get("user")),
                           password=str(self._creds.get("password")),)
        self.sftp = ssh_client.open_sftp()

    @retry_handler(exceptions=TimeoutError, total_tries=2, initial_wait=60)
    def _query_handler(self, params: dict) -> dict:
        """
        Fetch XML from url
        """
        self._connect()

        file_path_var = self.config["file_path"]
        file_path = file_path_var.format(dir_var=params.get('dir_var'),
                                         date_dir=params.get('date_dir'),
                                         date=params.get('date'))
        try:
            with self.sftp.open(file_path, "r") as file:
                file.seek(0, 0)
                data = json.load(file)

                return {
                        "data": data,
                        "dir_var": params.get('dir_var'),
                        "date": params.get('date').replace('-', '')
                    }
        except FileNotFoundError:
            print(f'No file called: {file_path}')
            return {
                "data": None,
                "dir_var": None,
                "date": None
            }

    def run_query(self) -> list:

        start_date: str = self.config["start_date"]
        end_date: str = self.config["end_date"]

        for dir_var in self.config["directory_variables"]:

            for date in date_range(start_date=start_date, end_date=end_date):
                date_dir = date_string_change_format(date_string=date,
                                                     output_format='%Y/%-m/%-d')
                yield self._query_handler(
                    params={
                        'dir_var': dir_var,
                        'date_dir': date_dir,
                        'date': date
                    }
                )



# pylint: disable=no-member
import sys

from google.cloud import bigquery
from google.oauth2 import service_account

from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.date_managers import date_range


class CustomGBQReader:
    """
    Custom Google BigQuery Console Reader
    """

    def __init__(self, creds_filepath=None, config_path=None, file_size_limit: int = 104857):

        self.credentials = load_file(creds_filepath, fmt="json")

        self.config = load_file(config_path, fmt="yml")
        self.file_size_limit = file_size_limit  # 5MB

    def get_auth(self):
        """
        Get our credentials initialised above and use those to get client

        """

        self.credentials['project_id'] = self.config.get('project_id')

        credentials = service_account.Credentials.from_service_account_info(
            self.credentials
        )
        client = bigquery.Client(
            credentials=credentials,
            project=credentials.project_id,
        )
        return client

    def build_query(self, date_run):
        """Build up query """
        # initialization
        project_id: str = self.config.get('project_id')
        table_id: str = self.config['table_id']
        query_prefix = f"SELECT *, NULL FROM {project_id}.{table_id}.events_{date_run} "
        query_suffix = " WHERE app_info.firebase_app_id = @firebase_id "
        query = query_prefix + query_suffix
        return query

    @staticmethod
    def _query_handler(client, query, params=None):
        """Query handler for the dataset"""

        result = None
        if params is not None:
            job_config = bigquery.QueryJobConfig()
            job_config.query_parameters = params

            query_job = client.query(query, location="US", job_config=job_config)
        # no parameters just a simple query
        else:
            query_job = client.query(query)
        try:
            result = query_job.result()
            return result
        except SyntaxError as err:
            if err.code == 400:
                print(f"Syntax error: {err}")
                sys.exit(1)
            else:
                print(f"Got unexpected error:\n {err}")
                sys.exit(1)

    def run_query(self):
        """
        Consumes a config file and loops through the dims
        to return relevant data from Google Big Query.
        """
        client = self.get_auth()

        start_date: str = self.config["start_date"]
        end_date: str = self.config["end_date"]

        print(f"Gathering data between given dates {start_date} and {end_date}. ")
        # split request by date to reduce 504 errors
        # BigQuery tables are named in format (project_id)_name_YYYYMMDD

        for firebase_id in self.config["firebase_ids"]:
            params = [
                bigquery.ScalarQueryParameter("firebase_id", "STRING", firebase_id),
            ]
            venture = self.config["firebase_ids"][firebase_id]

            for date in date_range(start_date=start_date, end_date=end_date):

                current_date = date.replace("-", "")
                query = self.build_query(current_date)

                print(f"Querying at date: {current_date}.")
                # run until none is returned or there is no more data in rows
                result = self._query_handler(client, query, params)

                current_array: list = []
                partition_counter: int = 0

                if result:
                    for partition_counter, row in enumerate(result):
                        current_array.append(dict(row.items()))
                        current_size = sys.getsizeof(current_array)
                        if current_size >= self.file_size_limit:
                            yield {
                                "date": date,
                                "venture": venture,
                                "data": current_array,
                                "partition": partition_counter,
                            }
                            current_array = []
                    if current_array:
                        yield {
                            "date": date,
                            "venture": venture,
                            "data": current_array,
                            "partition": partition_counter,
                        }



import urllib.parse
from datetime import datetime
import requests
from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.retry_managers import retry_handler


class WebVitalReader:
    """
    Custom FTP Reader
    """

    def __init__(self, creds_file, config_file):
        self._creds: dict = load_file(creds_file, "yml")

        self.config = load_file(config_file)

    @retry_handler(exceptions=TimeoutError, total_tries=5, initial_wait=900)
    def _query_handler(self) -> dict:
        """
        Fetch json from API
        """

        url = self.build_url()

        headers = {
            "Authorization": self._creds.get("bearer_token")
        }

        ts_now = int(datetime.timestamp(datetime.now()))

        return {
            "response": requests.get(url=url, headers=headers, timeout=30).json(),
            "venture": self.config.get("venture").get("name"),
            "ts_now": ts_now
        }

    def build_url(self) -> str:

        endpoint = self.config.get("endpoint")
        use_case = self.config.get("use_case")
        venture_url = self.config.get("venture").get("url")

        params_dict = self.config.get("url_params", "")
        params = urllib.parse.urlencode(params_dict)

        url = f"{endpoint}/{use_case}/{venture_url}"

        return url + f"?{params}" if params else url

    def run_query(self) -> dict:

        result = self._query_handler()

        return result



"""READER FILE FOR ZOHO RECRUIT"""
# pylint: disable=wrong-import-order,arguments-differ, broad-except, too-few-public-methods

from datetime import datetime
from typing import List, Dict, Union, Any, Generator

from sdc_dp_helpers.base_readers import BaseReader
from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.zoho_crm.zoho_crm_sdk import ZohoCRMAPI


class ZohoCRMReader(BaseReader):
    """Zoho CRMReader Class

    Args:
        BaseReader (_type_): API Reader Class
    """

    def __init__(self, creds_filepath: str, config_filepath: str):
        self.configs: Dict[Any, Any] = load_file(config_filepath)
        self.creds: Dict[Any, Any] = load_file(creds_filepath)
        self.service = self._get_auth()
        self.success: List[bool] = []

    def _get_auth(self):
        """Authenticate the service"""

        service = ZohoCRMAPI(creds=self.creds, configs=self.configs)
        return service

    def _query_handler(
        self,
    ) -> Generator[Dict[str, Union[str, List[Dict[Any, Any]]]], None, None]:
        """Method ot get the endpoint handler to use

        Args:
            config (Dict[Any, Any]): configs passed to the API endpoint handler

        Returns:
            List[Dict[Any, Any]]:  A list of dictionary of each response
        """
        results = self.service.fetch_data()
        for result in results:
            if result["data"]:
                yield {
                    "data": result["data"],
                    "module_partition": result["module_partition"],
                    "partition_counter": result["partition_counter"],
                    "date": datetime.now().strftime("%Y%m%d"),
                }
                self.is_success()
            else:
                self.not_success()

    def run_query(
        self,
    ) -> Generator[Dict[str, Union[str, List[Dict[Any, Any]]]], None, None]:
        """Method to run the query"""
        results = self._query_handler()

        yield from results



# pylint: disable=too-few-public-methods,unused-imports,line-too-long,too-many-arguments,arguments-differ

"""
    CUSTOM WRITER CLASS FOR XERO DATA
"""
# import os
import re
from typing import Union, List, Any, Dict, Tuple, Optional
from sdc_dp_helpers.base_writers import BaseWriter


class XeroWriter(BaseWriter):
    """
    Xero writer class
    """
    def __init__(
        self,
        bucket: str,
        folder_path: str,
        destination: str,
        configs: Optional[Dict[str, Any]] = None,
        clear_destination: bool = False
        ):
        self.current_destination: Union[str, None] = None
        super().__init__(
            bucket=bucket,
            folder_path=folder_path,
            destination=destination,
            configs=configs,
            clear_destination=clear_destination,
        )

    def verify_data(
        self, payload: Dict[str, Any]
    ) -> Tuple[str, Union[List[Dict[Any, Any]], Dict[Any, Any], Any]]:
        """verifies data before writting to destination also designs the destination file path

        Args:
            payload (Dict[str, Any]): expecting a dictionary having data, date and dimension

        Raises:
            KeyError: if we do not find the exact keys we expect from the payload
            TypeError: if provided data object is not a list

        Returns:
            Tuple[str, Union[List[Dict[Any, Any]], Dict[Any, Any], Any]]: full destination path and
                                                                          data to be written
        """
        # confirm the payload keys are matching accurately with what is expected
        if not {"data", "date", "endpoint", "trackingCategoryId", "tenant_name"}.issubset(set(payload.keys())):
            raise KeyError("invalid payload expecting: data, date, endpoint, trackingCategoryId, tenant_name")

        if not isinstance(payload["data"], list):
            raise TypeError("invalid data passed: expected List[Dict[Any,Any]]")

        if not re.search(r"\d{4}\-\d{2}\-\d{2}", payload["date"]):
            raise ValueError(
                "Invalid date format for partitioning expected: 'YYYY-mm-dd'"
            )

        _data: list = payload["data"]
        _date: str = payload["date"].replace("-", "")
        _endpoint: str= payload['endpoint']
        _tracking_categoryid=payload['trackingCategoryId']
        _tenant_name = payload['tenant_name']
        if _tracking_categoryid:
            write_path = f"{self.folder_path}/{_endpoint}/{_tenant_name}/{_tracking_categoryid}/{_date}"
        else:
            write_path=f"{self.folder_path}/{_endpoint}/{_tenant_name}/{_date}"
        return write_path, _data



"""MODULE FOR RECORDS ZOHO CRM DATA"""
# pylint: disable=arguments-differ,too-few-public-methods,inconsistent-return-statements,unused-argument,too-many-locals,broad-except,too-many-branches
import sys
import os
from abc import ABC, abstractmethod
from typing import Union, List, Dict, Any

from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
from zcrmsdk.src.com.zoho.crm.api.util.choice import Choice
from zcrmsdk.src.com.zoho.crm.api.users.user import User
from zcrmsdk.src.com.zoho.crm.api.layouts.layout import Layout
from zcrmsdk.src.com.zoho.crm.api.roles.role import Role
from zcrmsdk.src.com.zoho.crm.api.profiles.profile import Profile
from zcrmsdk.src.com.zoho.crm.api.record.record import Record

from zcrmsdk.src.com.zoho.crm.api.record import RecordOperations
from zcrmsdk.src.com.zoho.crm.api.record import (
    GetRecordsParam,
    GetRecordsHeader,
    APIException,
    ResponseWrapper,
)
from zcrmsdk.src.com.zoho.crm.api import HeaderMap, ParameterMap
from sdc_dp_helpers.api_utilities.retry_managers import retry_handler, request_handler
from sdc_dp_helpers.api_utilities.date_managers import date_string_handler


class InternalError(Exception):
    """
    Internal Server Error
    Resolution: Unexpected and unhandled exception in Server. Contact support team.
    """


class ErrorHandler:
    """class to handle errors"""

    def __init__(self):
        self.errors: list = []

    def append_error(self, module: str, error):
        """method to add errors to list"""
        self.errors.append({module: error})

    def get_errors(self):
        """method to return the error list"""
        return "\n".join([str(err) for err in self.errors])


class RequestHandler(ABC):
    """Base class to handle API calls"""

    @abstractmethod
    def api_call(self, **kwargs):
        """class method to make API calls"""
        raise NotImplementedError


class RecordsEndpointHandler:
    """Endpoint Handler for Records"""

    def __init__(self, configs: dict):
        self.configs = configs
        self.file_size_limit = self.configs.get("file_size_limit", 104857)
        self.error_handler = ErrorHandler()
        self.request_handler: RequestHandler = RecordsRequestHandler(
            error_handler=self.error_handler
        )

    def paginate(self, module_name: str, **kwargs):
        """Function to paginate through the whole dataset"""
        result_array: List[Dict[Any, Any]] = []
        current_size: int = 0
        more_records: bool = True
        page: int = self.configs.get("page", 1)
        page_limit: int = self.configs.get("page_limit")
        partition_counter: int = 0
        while more_records:
            print(f"fetching data for {module_name} page : {page}")
            response = self.request_handler.api_call(
                module_name=module_name, page=page, configs=self.configs
            )
            if response is None:
                yield {"data": [], "partition_counter": None, "module_partition": None}
                break

            request_info, results = response
            if request_info is not None:
                more_records: bool = request_info.get_more_records()

            for row in results["data"]:
                result_array.append(row)
                current_size += sys.getsizeof(row)
                if current_size >= self.file_size_limit:
                    yield {
                        "data": result_array,
                        "partition_counter": partition_counter,
                        "module_partition": results["module_partition"],
                    }
                    result_array = []
                    current_size: int = 0
                partition_counter += 1
            if result_array:
                yield {
                    "data": result_array,
                    "partition_counter": partition_counter,
                    "module_partition": results["module_partition"],
                }
            if page_limit is not None and page >= page_limit:
                print("early exit we have reached page limit")
                break

            page += 1


class RecordsRequestHandler(RequestHandler):
    """Requests Handler for Records Endpoint"""

    def __init__(self, error_handler):
        self.error_handler = error_handler

    @staticmethod
    def process_more_columns(column_name: str, value):
        """Method to process the data"""
        if isinstance(value, (User, Layout, Profile, Role)):
            return column_name, str(value.get_id())
        if isinstance(value, Choice):
            return column_name, value.get_value()
        if isinstance(value, Record):
            return column_name, value.get_id()
        return column_name, str(value)

    @staticmethod
    def build_query(configs: dict, cvid: Union[str, None], page: int, page_token=None):
        """build query to be run by the reader"""

        time_format = configs.get("time_format", "%Y-%m-%dT00:00:01+00:00")
        modified_since = date_string_handler(
            configs.get("modified_since", "50_years_ago"), time_format=time_format
        )
        param_instance = ParameterMap()
        param_instance.add(
            GetRecordsParam.converted, str(configs.get("converted", "both")).lower()
        )

        if page_token is None:
            param_instance.add(GetRecordsParam.page, page)
            param_instance.add(GetRecordsParam.per_page, configs["per_page"])
        if cvid is not None:
            print("adding cvid")
            param_instance.add(GetRecordsParam.cvid, str(cvid))
        if configs.get("sort_by") and cvid is None:
            param_instance.add(GetRecordsParam.sort_order, configs.get("sort_order"))
            param_instance.add(GetRecordsParam.sort_by, configs["sort_by"])
        param_instance.add(
            GetRecordsParam.include_child, str(configs["include_child"]).lower()
        )
        header_instance = HeaderMap()
        header_instance.add(
            GetRecordsHeader.if_modified_since,
            modified_since,  # configs["modified_since"]
        )
        print(modified_since)
        return param_instance, header_instance

    @retry_handler(
        exceptions=ConnectionError, total_tries=3, initial_wait=2, backoff_factor=2
    )
    @retry_handler(
        exceptions=InternalError, total_tries=3, initial_wait=2, backoff_factor=2
    )
    @request_handler(
        wait=int(os.environ.get("API_WAIT_TIME", 5)),
        backoff_factor=0.01,
        backoff_method="random",
    )
    def api_call(self, module_name: str, page: int, configs: dict, page_token=None):
        """calls the api call handler"""
        cvid = None
        request_info = None
        module_prefix = ""
        modulename_combination = module_name.rsplit("-", maxsplit=1)
        module_name = modulename_combination[0]
        if len(modulename_combination) > 1:
            module_prefix = modulename_combination[1]

        try:
            if configs.get("cvid") and module_prefix.startswith("CustomView"):
                cvid = configs["cvid"].get(f"{module_name}-{module_prefix}")

            param_instance, header_instance = self.build_query(
                configs=configs, cvid=cvid, page=page, page_token=page_token
            )

            response = RecordOperations().get_records(
                module_name, param_instance, header_instance
            )
            if (response is None) or (response.get_status_code() in [204, 304]):
                if response is not None:
                    print(
                        f"{module_name} returned No Content"
                        if response.get_status_code() == 204
                        else f"{module_name} returned Not Modified"
                    )
                return

            # Get object from response
            response_object = response.get_object()
            if response_object is None:
                return
            # Check if expected ResponseWrapper instance is received.
            if isinstance(response_object, ResponseWrapper):
                record_list = response_object.get_data()
                request_info = response_object.get_info()
                result_dict: Dict[Any, Any] = {}
                result_array: List[Dict[Any, Any]] = []
                for _, record in enumerate(record_list):
                    for key, value in record.get_key_values().items():
                        key, value = self.process_more_columns(key, value)
                        result_dict.update({key: value})
                    result_array.append(result_dict)
                    result_dict: Dict[Any, Any] = {}
                module_partition = f"{module_name}"
                if cvid is not None:
                    module_partition = f"{module_name}_{cvid}"
                return request_info, {
                    "module_partition": module_partition,
                    "data": result_array,
                }

            # Check if the request returned an exception
            if isinstance(response_object, APIException):
                error_code = response_object.get_code().get_value()
                error_details = response_object.get_details()
                self.error_handler.append_error(
                    module_name, f"{error_code}, {str(error_details)}"
                )
                if error_code == "INTERNAL_ERROR":
                    raise InternalError(
                        f"{module_name} {error_code}, {str(error_details)}"
                    )
            return

        except SDKException as exc:
            if "ConnectionError" in str(exc):
                raise ConnectionError(exc) from exc
            self.error_handler.append_error(module_name, exc.error_message)
        except Exception as exc:
            self.error_handler.append_error(module_name, exc.args[0])
        return



# pylint: disable=too-few-public-methods,disable=R0913, disable=C0301

"""
    CUSTOM WRITER CLASSES
"""
import re
import json
from typing import Union, List, Any, Dict, Tuple, Optional
import boto3

from sdc_dp_helpers.base_writers import BaseWriter


class FalconWriter(BaseWriter):
    """WRITER FOR FALCON DATA"""

    def __init__(
        self,
        bucket: str,
        folder_path: str,
        destination: str,
        configs: Optional[Dict[str, Any]] = None,
        clear_destination: bool = False,
    ):
        self.success: List[bool] = []
        self.current_destination: Union[str, None] = None
        self.boto3_session = boto3.Session()
        self.s3_resource = self.boto3_session.resource("s3")
        super().__init__(
            bucket=bucket,
            folder_path=folder_path,
            destination=destination,
            configs=configs,
            clear_destination=clear_destination,
        )

    def verify_data(
        self, payload: Dict[str, Any]
    ) -> Tuple[str, Union[List[Dict[Any, Any]], Dict[Any, Any], Any]]:
        """verifies data before writting to destination also designs the destination file path

        Args:
            payload (Dict[str, Any]): expecting a dictionary having data, date and dimension

        Raises:
            KeyError: if we do not find the exact keys we expect from the payload
            TypeError: if provided data object is not a list

        Returns:
            Tuple[str, Union[List[Dict[Any, Any]], Dict[Any, Any], Any]]: full destination path and
        """
        if not {"data", "date", "networks"}.issubset(set(payload.keys())):
            raise KeyError("Invalid payload expecting: date, data, networks")

        if not payload.get("date"):
            raise ValueError("Date cannot be None")
        if not re.search(r"\d{4}\-\d{2}\-\d{2}", payload["date"]):
            raise ValueError(
                "Invalid date format for partitioning expected: 'YYYY-mm-dd'"
            )
        if not isinstance(payload["data"], list):
            raise TypeError(
                "Invalid data value passed: expected List[Dict, Dict]")
        if not payload["networks"]:
            raise ValueError("Networks can not be None")
        if not isinstance(payload["networks"], str):
            raise TypeError("Invalid networks value passed: expected str")
        _date = payload.get("date").replace("-", "")
        _data: list = payload["data"]
        _networks: str = payload["networks"]
        _write_path: str = f"{self.folder_path}/{_networks}/{_date}"
        return _write_path, _data

    def write_stats_to_s3(
        self,
        stats: str,
        date: str,
        network: str,
        endpoint: str,
        target_table_name: str = "falcon_api_statistics"
    ) -> None:
        """
        Write statistics about the number of api calls on a particular date, network and endpoint.
        """
        write_path: str = f"{target_table_name}/date={date}/endpoint={endpoint}/network={network}/stats.json"
        if stats:
            print(
                f"Writing data to s3://{self.bucket}/{write_path} partitioned by networks and date."
            )
            self.s3_resource.Object(self.bucket, write_path).put(
                Body=json.dumps(stats))
            self.is_success()

    def write_channel_ids_s3(
        self,
        channel_ids,
        target_table_name: str = "falcon_channel_ids"
    ) -> None:
        """
        Write statistics about the number of api calls on a particular date, network and endpoint.
        """
        write_path: str = f"{target_table_name}/channel_ids.json"
        if channel_ids:
            print(
                f"Writing data to s3://{self.bucket}/{write_path} partitioned by networks and date."
            )
            self.s3_resource.Object(self.bucket, write_path).put(
                Body=json.dumps(channel_ids))
            self.is_success()



"""MODULE FOR FETCHING ZOHO CRM DATA"""
# pylint: disable=arguments-differ,too-few-public-methods,duplicate-code,broad-except,unused-variable

from pathlib import Path
from typing import Dict, Any
from zcrmsdk.src.com.zoho.api.authenticator.store import FileStore
from zcrmsdk.src.com.zoho.api.logger import Logger
from zcrmsdk.src.com.zoho.crm.api.user_signature import UserSignature
from zcrmsdk.src.com.zoho.crm.api.dc import USDataCenter
from zcrmsdk.src.com.zoho.api.authenticator.oauth_token import OAuthToken, TokenType
from zcrmsdk.src.com.zoho.crm.api.exception.sdk_exception import SDKException

# from zcrmsdk.src.com.zoho.crm.api.request_proxy import RequestProxy
from zcrmsdk.src.com.zoho.crm.api.sdk_config import SDKConfig
from zcrmsdk.src.com.zoho.crm.api.initializer import Initializer

from sdc_dp_helpers.zoho_crm.users_endpoint import UsersEndpointHandler
from sdc_dp_helpers.zoho_crm.records_endpoint import RecordsEndpointHandler
from sdc_dp_helpers.zoho_crm.bulk_endpoint import BulkEndpointHandler

from sdc_dp_helpers.api_utilities.retry_managers import retry_handler


class Authenticator:
    """Authenticator class"""

    def __init__(self, creds: dict):
        self.creds = creds
        self.current_path = Path(__file__).parent.resolve()

    @staticmethod
    def fetch_or_create_path(folder_path: str):
        """helper to ensure the path exists"""
        root_dir = Path(__file__).parent.resolve()
        if folder_path is not None or folder_path.strip() != ".":
            root_dir = f"{root_dir}/{folder_path}"

        Path(f"{root_dir}").mkdir(parents=True, exist_ok=True)
        return root_dir

    @retry_handler(
        exceptions=ConnectionError, total_tries=3, initial_wait=2, backoff_factor=2
    )
    def authenticate(self):
        """authenticator class method"""
        try:
            store = FileStore(
                file_path=f"{self.fetch_or_create_path('secrets/persistence')}/secrets.txt"
            )
            environment = USDataCenter.PRODUCTION()
            token = OAuthToken(
                client_id=self.creds["client_id"],
                client_secret=self.creds["client_secret"],
                token=self.creds["refresh_token"],
                token_type=TokenType.REFRESH,
                redirect_url=self.creds["redirect_uri"],
            )
            logger = Logger.get_instance(
                level=Logger.Levels.INFO,
                file_path=f"{self.fetch_or_create_path('logs')}/python_sdk_log.log",
            )
            # Create an UserSignature instance that takes user Email as parameter
            resource_path = self.fetch_or_create_path(".")
            user = UserSignature(email=self.creds["currentUserEmail"])
            config = SDKConfig(auto_refresh_fields=True, pick_list_validation=False)

            Initializer.initialize(
                user=user,
                environment=environment,
                token=token,
                store=store,
                sdk_config=config,
                resource_path=resource_path,
                logger=logger,
                # proxy=request_proxy
            )
        except SDKException as exc:
            print("Error at authentication")
            if "ConnectionError" in str(exc):
                raise ConnectionError(exc) from exc

        return store


class HandledExceptionsError(Exception):
    """class for exceptions that we handled during the api run"""


class ZohoCRMAPI:
    """Sample"""

    def __init__(
        self,
        creds: Dict[Any, Any],
        configs: Dict[Any, Any],
    ):
        self.creds = creds
        self.configs = configs
        self.authenticator = Authenticator(creds=self.creds)
        self.service = self.authenticator.authenticate()
        self.file_size_limit = self.configs.get("file_size_limit", 104857)

    def get_end_point_handler(self, end_point: str):
        """Gets the API Handler to use in getting data"""
        endpoints = {
            "bulk": BulkEndpointHandler,
            "records": RecordsEndpointHandler,
            "users": UsersEndpointHandler,
        }
        if end_point not in endpoints:
            raise KeyError(f"provided endpoint not yet implemented: {end_point}")

        return endpoints[end_point](configs=self.configs)

    def fetch_data(self):
        """method to fetch data"""
        end_point: str = self.configs["end_point"]
        endpoint_handler = self.get_end_point_handler(end_point)
        modules = (
            self.configs["modules"] if end_point != "users" else self.configs["type"]
        )
        for module_name in modules:
            response = endpoint_handler.paginate(module_name=module_name)
            yield from response
        if endpoint_handler.error_handler.errors:
            raise HandledExceptionsError(endpoint_handler.error_handler.get_errors())



"""CUSTOM MODULE TO WRITE GA DATA"""
# pylint: disable=too-few-public-methods, too-many-arguments
from typing import Union, List, Any, Dict, Tuple, Optional
from sdc_dp_helpers.base_writers import BaseWriter


class GAV3Writer(BaseWriter):
    """WRITER FOR GA V3 DATA"""

    def __init__(
        self,
        bucket: str,
        folder_path: str,
        destination: str,
        configs: Optional[Dict[str, Any]] = None,
        clear_destination: bool = False,
    ):
        self.success: List[bool] = []
        self.current_destination: Union[str, None] = None
        super().__init__(
            bucket=bucket,
            folder_path=folder_path,
            destination=destination,
            configs=configs,
            clear_destination=clear_destination,
        )

    def verify_data(
        self, payload: Dict[str, Any]
    ) -> Tuple[str, Union[List[Dict[Any, Any]], Dict[Any, Any], Any]]:
        """
        This pulls the yielded dataset from the GA reader in a manner
        that consumes the dataset of the given view_id and date,
        and writes it to s3 so that duplication does not occur.
        :param payload: This is a key value object that looks like:
                        {
                            "data": list(),
                            "date": string,
                            "view_id": string
                        }
        """
        if list(payload.keys()).sort() != ["data", "date", "view_id"].sort():
            raise KeyError("Invalid payload")
        if not isinstance(payload["data"], list):
            raise TypeError("invalid data passed: expected List[Dict[Any,Any]]")
        if not isinstance(payload["view_id"], str):
            raise TypeError("invalid view_id passed: expected string")
        _view_id= payload.get("view_id")
        _date = payload.get("date")
        if _date:
            _date = _date.replace("-", "")
        else:
            raise KeyError("Date cannot be None")
        _data = payload.get("data")
        write_path: str = f"{self.folder_path}/{_view_id}/{_date}"

        return write_path, _data



"""
    CUSTOM WRITER CLASSES
        - Class which manages writer tasks like
        auth, write metadata, write file, create dir structure
"""
import json
import boto3


class SFTPWriter:
    """SFTPWriter Class"""

    def __init__(self, bucket, folder_path, profile_name=None):
        if profile_name is None:
            self.boto3_session = boto3.Session()
        else:
            self.boto3_session = boto3.Session(profile_name=profile_name)
        self.s3_resource = self.boto3_session.resource("s3")
        self.bucket = bucket
        self.folder_path = folder_path

    def write_to_s3(self, payload):
        """
        Writes json to specified s3 location.
        """

        data = json.dumps(payload['data']).encode('utf-8')
        dir_var = payload['dir_var']
        date = payload['date']

        write_path = f"{self.folder_path}/{date}/{dir_var}.json"
        print(f"Writing data to s3://{self.bucket}/{write_path}")
        self.s3_resource.Object(self.bucket, write_path).put(Body=data)



"""MODULE FOR USERS ZOHO CRM DATA"""
# pylint: disable=arguments-differ,too-few-public-methods,inconsistent-return-statements,unused-argument,too-many-locals,broad-except

import sys
import os
from abc import ABC, abstractmethod
from typing import List, Dict, Any

from zcrmsdk.src.com.zoho.crm.api.exception import SDKException

from zcrmsdk.src.com.zoho.crm.api.util.choice import Choice
from zcrmsdk.src.com.zoho.crm.api.users.user import User
from zcrmsdk.src.com.zoho.crm.api.layouts.layout import Layout
from zcrmsdk.src.com.zoho.crm.api.record.record import Record
from zcrmsdk.src.com.zoho.crm.api.roles.role import Role
from zcrmsdk.src.com.zoho.crm.api.profiles.profile import Profile

from zcrmsdk.src.com.zoho.crm.api import HeaderMap, ParameterMap
from zcrmsdk.src.com.zoho.crm.api.users import (
    UsersOperations,
    GetUsersParam,
    GetUsersHeader,
    APIException,
    ResponseWrapper,
)
from sdc_dp_helpers.api_utilities.retry_managers import retry_handler, request_handler
from sdc_dp_helpers.api_utilities.date_managers import date_string_handler


class InternalError(Exception):
    """
    Internal Server Error
    Resolution: Unexpected and unhandled exception in Server. Contact support team.
    """


class ErrorHandler:
    """class to handle errors"""

    def __init__(self):
        self.errors: list = []

    def append_error(self, module: str, error):
        """method to add errors to list"""
        self.errors.append({module: error})

    def get_errors(self):
        """method to return the error list"""
        return "\n".join([str(err) for err in self.errors])


class RequestHandler(ABC):
    """Base class to handle API calls"""

    @abstractmethod
    def api_call(self, **kwargs):
        """class method to make API calls"""
        raise NotImplementedError


class UsersEndpointHandler:
    """crm users endpoint api handler"""

    def __init__(self, configs: dict):
        self.configs = configs
        self.file_size_limit = self.configs.get("file_size_limit", 104857)
        self.error_handler = ErrorHandler()
        self.request_handler: RequestHandler = UsersRequestHandler(
            error_handler=self.error_handler
        )

    def paginate(self, module_name: str, **kwargs):
        """Function to paginate through the whole dataset"""
        result_array: List[Dict[Any, Any]] = []
        current_size: int = 0
        more_records: bool = True
        page: int = self.configs.get("page", 1)
        page_limit: int = self.configs.get("page_limit")
        partition_counter: int = 0
        while more_records:
            print(f"fetching data for {module_name} page : {page}")
            response = self.request_handler.api_call(
                module_name=module_name, page=page, configs=self.configs
            )
            if response is None:
                yield {"data": [], "partition_counter": None, "module_partition": None}
                break

            request_info, results = response
            if request_info is not None:
                more_records: bool = request_info.get_more_records()

            for row in results["data"]:
                result_array.append(row)
                current_size += sys.getsizeof(row)
                if current_size >= self.file_size_limit:
                    yield {
                        "data": result_array,
                        "partition_counter": partition_counter,
                        "module_partition": results["module_partition"],
                    }
                    result_array = []
                    current_size: int = 0
                partition_counter += 1
            if result_array:
                yield {
                    "data": result_array,
                    "partition_counter": partition_counter,
                    "module_partition": results["module_partition"],
                }
            if page_limit is not None and page >= page_limit:
                print("early exit we have reached page limit")
                break
            page += 1


class UsersRequestHandler(RequestHandler):
    """Request Handler to Get Users"""

    def __init__(self, error_handler):
        self.error_handler = error_handler

    @staticmethod
    def build_query(configs: dict, module_name: str, page: int):
        """build query to be run by the reader"""

        time_format = configs.get("time_format", "%Y-%m-%dT00:00:01+00:00")
        modified_since = date_string_handler(
            configs.get("modified_since", "50_years_ago"), time_format=time_format
        )
        param_instance = ParameterMap()
        param_instance.add(GetUsersParam.type, module_name)
        param_instance.add(GetUsersParam.page, page)
        param_instance.add(GetUsersParam.per_page, configs.get("per_page", 200))
        header_instance = HeaderMap()
        header_instance.add(GetUsersHeader.if_modified_since, modified_since)
        return param_instance, header_instance

    @staticmethod
    def process_more_columns(column_name: str, value):
        """Method to process the data"""
        if isinstance(value, (User, Layout, Profile, Role)):
            return column_name, str(value.get_id())
        if isinstance(value, Choice):
            return column_name, value.get_value()
        if isinstance(value, Record):
            return column_name, value.get_id()
        return column_name, str(value)

    @retry_handler(
        exceptions=ConnectionError, total_tries=3, initial_wait=2, backoff_factor=2
    )
    @retry_handler(
        exceptions=InternalError, total_tries=3, initial_wait=2, backoff_factor=2
    )
    @request_handler(
        wait=int(os.environ.get("API_WAIT_TIME", 5)),
        backoff_factor=0.01,
        backoff_method="random",
    )
    def api_call(self, module_name: str, page: int, configs: dict, **kwargs):
        """calls the api call handler"""
        request_info = None

        try:
            param_instance, header_instance = self.build_query(
                configs=configs, module_name=module_name, page=page
            )
            response = UsersOperations().get_users(param_instance, header_instance)
            if (response is None) or (response.get_status_code() in [204, 304]):
                if response is not None:
                    print(
                        f"{module_name} returned No Content"
                        if response.get_status_code() == 204
                        else f"{module_name} returned Not Modified"
                    )
                return

            # Get object from response
            response_object = response.get_object()
            if response_object is None:
                print("response get_object is none")
                return
            # Check if expected ResponseWrapper instance is received.
            if isinstance(response_object, ResponseWrapper):
                print("In response wrapper")
                record_list = response_object.get_users()
                request_info = response_object.get_info()
                result_dict: Dict[Any, Any] = {}
                result_array: List[Dict[Any, Any]] = []
                for _, record in enumerate(record_list):
                    for key, value in record.get_key_values().items():
                        key, value = self.process_more_columns(key, value)
                        result_dict.update({key: value})
                    result_array.append(result_dict)
                    result_dict: Dict[Any, Any] = {}
                module_partition = f"{module_name}"
                return request_info, {
                    "module_partition": module_partition,
                    "data": result_array,
                }

            # Check if the request returned an exception
            if isinstance(response_object, APIException):
                error_code = response_object.get_code().get_value()
                error_details = response_object.get_details()
                self.error_handler.append_error(
                    module_name, f"{error_code}, {str(error_details)}"
                )
                if error_code == "INTERNAL_ERROR":
                    raise InternalError(
                        f"{module_name} {error_code}, {str(error_details)}"
                    )
            return

        except SDKException as exc:
            if "ConnectionError" in str(exc):
                raise ConnectionError(exc) from exc
            self.error_handler.append_error(module_name, exc.error_message)
        except Exception as exc:
            self.error_handler.append_error(module_name, exc.args[0])
        return



"""BASE WRITER MODULE"""

# pylint: disable=too-few-public-methods, no-member, too-many-arguments
from abc import ABC, abstractmethod
import json
import csv
import gzip
import time
from io import BytesIO
from pathlib import Path
from typing import Union, Any, List, Dict, Optional, Tuple
import boto3
from googleapiclient import discovery
from google.oauth2 import service_account


class Authenticator(ABC):
    """Interface for Resource Authentication"""

    def authenticate(self, configs: Optional[Dict[str, Any]]):
        """
        class method that does the actual authentication

        Raises:
            NotImplementedError: method must be overriden by subclasses

        Returns:
            None : has no return value
        """
        print(configs)
        return True


class BaseResourceWriter(ABC):
    """Interface class for resource writers"""

    @abstractmethod
    def write_to_destination(self, write_path: str, data: Any) -> None:
        """Method to write data to final destination resource

        Args:
            write_path (str): string of the path to write the data
            data (Any): the data object to be written

        Raises:
            NotImplementedError: should be implemented by the child classes

        Returns:
            None : has no return
        """
        raise NotImplementedError

    def delete_destination(self, delete_path: str) -> None:
        """Method to delete from the destination

        Args:
            delete_path (str): path to delete

        Raises:
            KeyError: if givne delete path does not exist
            exc: _description_
            NotImplementedError: _description_
            NotImplementedError: _description_

        Returns:
            Union[Error|PrintMessage]: Either raises KeyError or print results
        """
        print(f"Not yet implemented to delete {delete_path}")


class AWSProfileAuthenticator(Authenticator):
    """Authenticator class for AWS Using AWS Profile"""

    def authenticate(self, configs: Optional[Dict[str, Any]] = None):
        """Authentication into AWS S3"""
        if configs is None:
            configs = {"profile_name": None}
        if configs.get("profile_name") is not None:
            boto3_session = boto3.Session(profile_name=configs["profile_name"])
        else:
            boto3_session = boto3.Session()

        return boto3_session.resource("s3")


class GCPServiceAccountAuthenticator(Authenticator):
    """Ayuthenticator class for GCP Using Service Account"""

    def authenticate(self, configs: Dict[str, Any]):
        """Authentication into Google Cloud"""
        if configs.get("service_account_info") is not None:
            credentials = service_account.Credentials.from_service_account_info(
                info=configs["service_account_info"]
            )
        elif configs.get("service_account_file") is not None:
            credentials = service_account.Credentials.from_service_account_file(
                filename=configs["service_account_file"]
            )
        else:
            raise KeyError(
                "missing service account file or info (dict) for authentication"
            )

        service = discovery.build("storage", "v1", credentials=credentials)
        return service


class LocalAuthenticator(Authenticator):
    """Placeholder classs for Local Authentication"""

    def authenticate(self, configs: Optional[Dict[str, Any]]):
        return True


class AWSS3JSONWriter(BaseResourceWriter):
    """Base Writer Class for S3 Destination"""

    def __init__(
        self,
        bucket: str,
        configs: Dict[str, Any],
        authenticator=AWSProfileAuthenticator,
    ):

        self.bucket: str = bucket
        self.configs: Optional[Dict[str, Any]] = configs
        self.resource = authenticator().authenticate(self.configs)

    def write_to_destination(self, write_path: str, data: Any) -> None:

        """Writes Data To JSON in S3 Generally expects a List, Dictionary
        Args:
            payload (Union[List[Dict[Any, Any]], Dict[Any, Any], Any]): can be any python data type

        Retruns:
            Tuple(Str,List[Dict[Any,Any]]): write_path full s3 path of object to be written
                                            data a list of dictionaries having the data
        """

        write_path = f"{write_path}.json"
        self.resource.Object(self.bucket, write_path).put(
            Body=json.dumps(data, indent=2)
        )
        print(f"done writting data to s3://{self.bucket}/{write_path}")

    def delete_destination(self, delete_path: str) -> None:
        """Method to delete from the destination

        Args:
            delete_path (str): path to delete

        Raises:
            KeyError: if givne delete path does not exist
        Returns:
            Union[Error|PrintMessage]: Either raises KeyError or print results
        """
        bucket = self.resource.Bucket(self.bucket)
        try:
            print(
                "5 seconds before you delete all records in the path\n\n"
                f"s3://{self.bucket}/{delete_path}\n\n"
            )
            time.sleep(5)
            deletes = bucket.objects.filter(Prefix=delete_path).delete()
            files = [row["Key"] for item in deletes for row in item["Deleted"]]
            print(f"deleted records are \n{files}\n")
        except Exception as exc:
            raise exc


class GCPCloudStorageJSONWriter(BaseResourceWriter):
    """Base Writer Class for GCP Cloud Storage Destination"""

    def __init__(
        self,
        bucket: str,
        configs: Dict[str, Any],
        authenticator=GCPServiceAccountAuthenticator,
    ):

        self.bucket: str = bucket
        self.configs: Dict[str, Any] = configs
        self.resource = authenticator().authenticate(self.configs)

    def write_to_destination(self, write_path: str, data: Any) -> None:

        """Writes Data To JSON in GCP Cloud Storage Generally expects a List, Dictionary"""

        write_path = f"{write_path}.json"
        self.resource.objects().insert(
            bucket=self.bucket, body=data, media_body=write_path
        ).execute()
        print(f"done writting data to gs://{self.bucket}/{write_path}")


class AWSS3JSONGZIPWriter(BaseResourceWriter):

    """AWS GZIP Writer"""

    def __init__(
        self,
        bucket: str,
        configs: Optional[Dict[str, Any]] = None,
        authenticator=AWSProfileAuthenticator,
    ):
        self.bucket: str = bucket
        self.configs = configs
        self.resource = authenticator().authenticate(self.configs)

    def write_to_destination(self, write_path: str, data: Any) -> None:
        gz_body = BytesIO()
        gz_file = gzip.GzipFile(None, "wb", 7, gz_body)
        gz_file.write(json.dumps(data).encode("utf-8"))
        gz_file.close()

        write_path = f"{write_path}.json.gz"

        self.resource.Bucket(self.bucket).put_object(
            Key=write_path,
            ContentEncoding="gzip",
            Body=gz_body.getvalue(),
        )
        print(f"done writting data to s3://{self.bucket}/{write_path}")

    def delete_destination(self, delete_path: str) -> None:
        """Method to delete from the destination

        Args:
            delete_path (str): path to delete

        Raises:
            KeyError: if givne delete path does not exist
        Returns:
            Union[Error|PrintMessage]: Either raises KeyError or print results
        """
        bucket = self.resource.Bucket(self.bucket)
        try:
            print(
                "5 seconds before you delete all records in the path\n\n"
                f"s3://{self.bucket}/{delete_path}\n\n"
            )
            time.sleep(5)
            deletes = bucket.objects.filter(Prefix=delete_path).delete()
            files = [row["Key"] for item in deletes for row in item["Deleted"]]
            print(f"deleted records are \n{files}\n")
        except Exception as exc:
            raise exc


class LocalJSONWriter(BaseResourceWriter):
    """Writer Class for Local Directories"""

    def __init__(
        self,
        bucket: str,
        configs: Optional[Dict[str, Any]] = None,
        authenticator=LocalAuthenticator,
    ):
        self.bucket: str = bucket
        self.configs = configs
        self.resource = authenticator().authenticate(self.configs)

    def write_to_destination(self, write_path: str, data: Any) -> None:
        full_path = write_path.rsplit("/", maxsplit=1)[0]
        Path(f"{self.bucket}/{full_path}").mkdir(parents=True, exist_ok=True)

        write_path = f"{write_path}.json"

        with open(
            f"{self.bucket}/{write_path}", mode="w", encoding="utf8"
        ) as dest_file:
            json.dump(data, dest_file, indent=4)
            print(f"done writting data to {self.bucket}/{write_path}")

    def delete_destination(self, delete_path: str) -> None:
        delete_path = f"{self.bucket}/{delete_path}"
        files = Path(f"{delete_path}").glob("*.*")
        files = list(files)
        print("5 seconds before you delete all records in the path\n" f"{delete_path}")
        time.sleep(5)
        _ = [file.unlink(missing_ok=True) for file in files]


class LocalCSVWriter(BaseResourceWriter):
    """Writer Class for Local CSV"""

    def __init__(
        self,
        bucket: str,
        configs: Optional[Dict[str, Any]] = None,
        authenticator=LocalAuthenticator,
    ):
        self.bucket: str = bucket
        self.configs = configs
        self.resource = authenticator().authenticate(self.configs)

    def write_to_destination(self, write_path: str, data: Any) -> None:
        full_path = write_path.rsplit("/", maxsplit=1)[0]
        Path(f"{self.bucket}/{full_path}").mkdir(parents=True, exist_ok=True)

        write_path = f"{write_path}.csv"
        field_names = list(data[0].keys())

        with open(f"{self.bucket}/{write_path}", mode="w", encoding="utf8") as myfile:
            writer = csv.DictWriter(
                myfile, fieldnames=field_names, quoting=csv.QUOTE_ALL
            )
            writer.writeheader()
            writer.writerows(data)
            print(f"done writting data to {self.bucket}/{write_path}")


class BaseWriter(ABC):
    """ "Normal Base Writer Class"""

    success: List[bool] = []

    def __init__(
        self,
        bucket: str,
        folder_path: str,
        destination: str,
        configs: Optional[Dict[str, Any]] = None,
        clear_destination: bool = False,
    ):
        self.bucket = bucket
        self.folder_path = folder_path
        self.destination = destination
        self.configs = configs
        self.resource = self._get_resource()
        self.clear_destination: bool = clear_destination
        self.current_destination: Union[str, None] = None

    def _get_resource(self) -> BaseResourceWriter:
        """Gets the appropriate WriterResource"""
        resource_writers = {
            "aws_s3_json": AWSS3JSONWriter,
            "gcp_cloudstorage_json": GCPCloudStorageJSONWriter,
            "local_json": LocalJSONWriter,
            "local_csv": LocalCSVWriter,
            "aws_s3_json_gzip": AWSS3JSONGZIPWriter,
        }
        if self.destination not in resource_writers:
            raise NotImplementedError(
                f"writer destination is wrong! allowed values: \n{list(resource_writers.keys())}"
            )
        resource = resource_writers[self.destination](self.bucket, self.configs)
        return resource

    def write_data(self, payload: Any):
        """Base Write Method to Destination"""
        write_path, data = self.verify_data(payload)
        delete_path = write_path.rsplit("/", maxsplit=1)[0].strip()

        if not data:
            self.not_success()
            return

        if (self.clear_destination is True) and (
            self.current_destination != delete_path
        ):
            self.resource.delete_destination(delete_path=delete_path)
            self.current_destination = delete_path
        self.resource.write_to_destination(write_path, data)
        self.is_success()

    def is_success(self) -> None:
        """Append True to the Success List Object"""
        self.success.append(True)

    def not_success(self) -> None:
        """Append False to the Success List Object"""
        self.success.append(False)

    @abstractmethod
    def verify_data(
        self, payload: Union[List[Dict[Any, Any]], Dict[Any, Any], Any]
    ) -> Tuple[str, Union[List[Dict[Any, Any]], Dict[Any, Any], Any]]:

        """Used to verify data is properly formatted and we have expected fields

        Args:
            payload (Union[List[Dict[Any, Any]], Dict[Any, Any], Any]): can be any python data type

        Raises:
            NotImplementedError: _description_

        Returns:
            Tuple[str, Union[List[Dict[Any, Any]], Dict[Any, Any], Any]]: write_path and data itself
        """
        raise NotImplementedError



"""
    CUSTOM READER CLASSES
        - Class which manages writer tasks like
        auth, write metadata, write file, create dir structure
"""
# pylint: disable=no-member,too-many-locals,broad-except,too-few-public-methods,arguments-differ,line-too-long,broad-exception-raised,inconsistent-return-statements, unused-import
import os
from datetime import datetime
import requests

from google.oauth2 import service_account
from google.protobuf.json_format import MessageToDict

from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.data_managers import multiple_regex_replace
from sdc_dp_helpers.api_utilities.retry_managers import retry_handler, request_handler
from sdc_dp_helpers.base_readers import BaseReader

class GoogleKnowledgeGraphReader(BaseReader):
    """v1  GoogleKnowledgeGraphReader Class"""
    def __init__(self, secrets_filepath: str, config_filepath: str):
        self.secrets: dict = load_file(secrets_filepath)
        self.config: dict = load_file(config_filepath)
        self.request_session: requests.Session = requests.Session()
        self.dataset: list = []
        self.success = []

    def query_partition(self, query: str, language: str = None) -> str:
        """Helper to get the query partition string
        :query: "string",
        :language:"string"}
        :returns: query_partition - str
        """
        query = query.strip()
        query_partition = multiple_regex_replace(
            query, {".": "_dot_", " ": "_space_", "+": "_plus_"}
        )
        if language:
            query_partition = f"{query_partition}_lang_{language}"

        return query_partition

    @request_handler(
        wait=int(os.environ.get("REQUEST_WAIT_TIME", 2)),
        backoff_factor=float(os.environ.get("REQUEST_BACKOFF_FACTOR", 0.01)),
        backoff_method=os.environ.get("REQUEST_BACKOFF_METHOD", "0.01"),
    )
    @retry_handler(
        exceptions=ConnectionError,
        total_tries=int(os.environ.get("TOTAL_RETRIES", 5)),
        initial_wait=float(os.environ.get("INITIAL_WAIT", 200)),
    )
    def _query_handler(self, query: str, language: str) -> dict:
        """Does the actual API call"""
        response_json: dict = {}
        if self.secrets.get("api_key") is None:
            raise KeyError("api_key missing in your secrets file")

        base_url = "https://kgsearch.googleapis.com/v1/entities:search"
        try:
            response = self.request_session.get(
                url=base_url,
                params={
                    "query": query,
                    "limit": self.config.get("limit", 1),
                    "indent": self.config.get("indent", True),
                    "key": self.secrets["api_key"],
                    "languages": language,
                },
            )
            status_code = response.status_code
            reason = response.reason
            if status_code in [404, 500, 202]:
                raise ConnectionError(
                    f"Google Knowledge Graph API failed. "
                    f"Status code: {status_code}, Reason: {reason}."
                )
            if status_code == 200:
                response_json = response.json()
            return response_json
        except Exception as err:
            raise Exception(f"Unexpected error: {err}") from err

    def run_query(self):
        """Handles the query results"""
        date = datetime.strftime(datetime.now(), "%Y-%m-%d")
        results = []
        for query, languages in self.config["queries"].items():
            for language in languages:
                payload: dict = self._query_handler(query, language)
                if payload:
                    # add metadata of date, query and language
                    query_partition: str = self.query_partition(query, language)
                    payload.update(
                        {
                            "date": date,
                            "query": query,
                            "language": language,
                            "query_partition": query_partition,
                        }
                    )
                    results.append(payload)
                    self.is_success()
                    print(
                        f"Got data for query:'{query}' language:'{language}' on '{date}'"
                    )
                if not payload:
                    self.not_success()
                    print(
                        f"No data for query:'{query}' language:'{language}' on '{date}'"
                    )
        # only return if we have daa in the results list
        if results:
            return {"date": date, "data": results}
        self.dataset = results
        return



# pylint: disable=no-member,inconsistent-return-statements,wrong-import-order,broad-exception-raised,no-else-return,arguments-differ
from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.xero.xero_sdk import XeroAPI
from sdc_dp_helpers.base_readers import BaseReader


class XeroReader(BaseReader):
    """
        Xero reader
    """
    def __init__(self, config_path: str, creds_filepath: str):
        self.creds_path: str = creds_filepath
        self.config: dict = load_file(config_path, "yml")
        self.service = self._get_auth()

    def _get_auth(self):
        self.service = XeroAPI(
            config=self.config, creds_filepath=self.creds_path
        )
        return self.service

    def _query_handler(self):
        if self.config.get("modules"):
            for module in self.config.get("modules", []):
                data = self.service.fetch_modules(module), module
        if self.config.get("reports"):
            for report in self.config.get("reports", []):
                data = self.service.fetch_reports(report), report
        return data

    def run_query(self):
        payload, endpoint_name = self._query_handler()
        if payload:
            self.is_success()
            return payload
        else:
            self.not_success()
            print(f"No data for endpoint {endpoint_name}")



"""
CUSTOM READERS MODULE FOR GOOGLE POSTMASTER READER
"""
from typing import Dict, List, Any, Union, Generator
from googleapiclient import discovery, errors
from google.oauth2 import service_account

from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.date_managers import (
    date_string_handler,
    date_range_iterator,
)
from sdc_dp_helpers.base_readers import BaseReader


class GooglePostmasterReader(BaseReader):
    """
    Google Postmaster Reader
    """

    def __init__(self, creds_filepath: str, config_filepath: str) -> None:
        super().__init__()
        self.secrets: Dict[Any, Any] = load_file(creds_filepath)
        self.config: Dict[Any, Any] = load_file(config_filepath)
        self.service = self._get_auth()
        self.success: List[bool] = []

    def _get_auth(self):
        """
        Get our credentials initialised above and use those to get client.
        """
        service = None
        try:
            credentials = service_account.Credentials.from_service_account_info(
                info=self.secrets,
                scopes=["https://www.googleapis.com/auth/postmaster.readonly"],
            )
            service = discovery.build(
                serviceName="gmailpostmastertools",
                version="v1beta1",
                credentials=credentials,
            )
            if service is None:
                raise RuntimeError("Service is null.")
            if not hasattr(service, "domains"):
                raise KeyError("Service does not have domains atribute.")
        except KeyError as err:
            self.not_success()
            print(f"An error has occurred: {err}")
        except RuntimeError as err:
            self.not_success()
            print(f"An error has occurred: {err}")
        return service

    def _query_handler(self, *args, **kwargs) -> dict:
        # pylint: disable=no-member
        """Handles the Query call"""
        response = None
        if not {"domain", "date"}.issubset(set(kwargs.keys())):
            raise KeyError("Invalid arguments - expecting: domain, date")
        response = (
            self.service.domains()
            .trafficStats()
            .get(name=f"domains/{kwargs['domain']}/trafficStats/{kwargs['date']}")
            .execute()
        )
        if response is None:
            raise RuntimeError("response is 'None'")
        self.is_success()
        return response

    def run_query(
        self,
    ) -> Union[
        Generator[Dict[List[Dict[Any, Any]], Any], None, None],
        Dict[List[Dict[Any, Any]], Any],
    ]:
        """Runs the query"""
        start_date = date_string_handler(self.config["start_date"])
        end_date = date_string_handler(self.config["end_date"])
        print(f" {start_date} {end_date} ")
        if start_date > end_date:
            raise ValueError(
                "An error has occured: "
                + f"- Start Date {start_date}"
                + f"is not before {end_date}."
            )
        if self.service is None:
            raise ValueError("Service is null.")
        domain = self.config["domain"]
        for date, _ in date_range_iterator(
            start_date=start_date,
            end_date=end_date,
            interval="1_day",
            end_inclusive=False,
            time_format="%Y%m%d",
        ):
            try:
                print(f"date : {date}")
                dataset = self._query_handler(domain=domain, date=date)
                yield {
                    "date": date,
                    "brand": domain.replace(".", "_"),
                    "data": [dataset],
                }
            except RuntimeError as err:
                self.not_success()
                print(f"Error occured: {err}")
            except errors.HttpError as err:
                self.not_success()
                print(f"An error occurred: {err}")
            except KeyError as err:
                self.not_success()
                print(f"An error occurred: {err}")



"""SDK MODULE FOR ONESIGNAL"""
# pylint: disable=arguments-differ,too-few-public-methods,too-many-arguments
import os
import sys
from abc import ABC, abstractmethod
from typing import Dict, Any, List, Tuple, Generator, Union
from datetime import datetime
import requests
from sdc_dp_helpers.api_utilities.retry_managers import request_handler, retry_handler
from sdc_dp_helpers.api_utilities.file_managers import parse_zip_to_csv
from sdc_dp_helpers.api_utilities.date_managers import date_string_handler


class AuthenticationError(Exception):
    """class for Authentication Errors"""


class HTTP500InternalServerError(Exception):
    """class to handle Internal Server Error 500"""


class NoDownloadYetError(Exception):
    """class for handling no download ready yet"""


class ExistingScheduledDownloadError(Exception):
    """Error raised if there exists a scheduled download"""


class CSVExportCreationError(Exception):
    """Error raised when we fail to create bulk job and get back a None"""


class ViewNotificationsDownloadError(Exception):
    """Error for view notifications data download"""


class APICallHandler(ABC):
    """Base class for API Calls"""

    def __init__(self, creds: dict):
        self.creds = creds
        self._header = {
            "accept": "application/json",
            "content-type": "application/json",
            "Authorization": f"Basic {self.creds['api_key']}",
            "User-Agent": "Mozilla/5.0",
        }

    @abstractmethod
    def api_call(self, **kwargs):
        """_summary_
        Raises:
            NotImplementedError: needs to be implemented for all child classes
        """
        raise NotImplementedError


class CreateBulkDownload(APICallHandler):
    """Class for Creating Bulk Doanload"""

    @retry_handler(
        HTTP500InternalServerError, total_tries=3, initial_wait=61, backoff_factor=1.2
    )
    @retry_handler(ConnectionError, total_tries=3, initial_wait=32, backoff_factor=1.2)
    @retry_handler(
        ExistingScheduledDownloadError, total_tries=10, initial_wait=2, backoff_factor=2
    )
    def api_call(self, session: requests.Session) -> Dict[str, str]:
        _app_id: str = self.creds["app_id"]
        url = f"https://onesignal.com/api/v1/players/csv_export?app_id={_app_id}"
        try:
            response = session.post(url=url, headers=self._header, timeout=61)
            response.raise_for_status()
            return response.json()

        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.Timeout,
            requests.exceptions.ChunkedEncodingError,
            requests.exceptions.ConnectTimeout,
        ) as exc:
            raise ConnectionError(exc) from exc

        except requests.exceptions.HTTPError as exc:
            raise HTTP500InternalServerError(exc) from exc

        except requests.exceptions.RequestException as exc:
            if "Please include a case-sensitive header of Authorization" in str(
                response.json()
            ):
                raise AuthenticationError(
                    "Please check the credentials used to make api call"
                ) from exc
            if "User already running another CSV export." in str(response.json()):
                print("Got into the down scheduled part")
                raise ExistingScheduledDownloadError(response.json()) from exc

            raise exc


class FetchBulkDownload(APICallHandler):
    """Class to fetch the generated bulk download"""

    @retry_handler(
        HTTP500InternalServerError, total_tries=3, initial_wait=61, backoff_factor=1.2
    )
    # we wait maximum 30 minutes for download to be ready otherwise we just raise an error
    @retry_handler(NoDownloadYetError, total_tries=10, initial_wait=2, backoff_factor=2)
    @retry_handler(ConnectionError, total_tries=3, initial_wait=32, backoff_factor=2)
    def api_call(self, session: requests.Session, url: str):
        """call for downloading the prepared csv export"""
        try:
            response = session.get(url=url, timeout=61)
            response.raise_for_status()
            return response.content
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.Timeout,
            requests.exceptions.ChunkedEncodingError,
            requests.exceptions.ConnectTimeout,
        ) as exc:
            raise ConnectionError(exc) from exc

        except requests.exceptions.HTTPError as exc:
            if response.status_code == 404:
                raise NoDownloadYetError(exc) from exc
            raise HTTP500InternalServerError(exc) from exc

        except requests.exceptions.RequestException as exc:
            if response.status_code == 404:
                raise NoDownloadYetError(exc) from exc
            raise exc


class ViewNotificationsDownload(APICallHandler):
    """class for fetching view notifications"""

    @retry_handler(
        HTTP500InternalServerError, total_tries=3, initial_wait=61, backoff_factor=1.2
    )
    @retry_handler(ConnectionError, total_tries=3, initial_wait=61, backoff_factor=1.2)
    @request_handler(
        wait=int(os.environ.get("API_WAIT_TIME", 5)),
        backoff_factor=0.01,
        backoff_method="random",
    )
    def api_call(
        self, session: requests.Session, limit: int, offset: int
    ) -> Tuple[bool, Dict[Any, Any]]:
        """
        Handles the view notification request attempt.
        """
        _app_id: str = self.creds["app_id"]
        url = f"https://onesignal.com/api/v1/notifications?app_id={_app_id}"
        more_records = False
        try:
            response = session.get(
                url=url,
                headers=self._header,
                params={"limit": limit, "total_count": "true", "offset": offset},
                timeout=61,
            )
            response.raise_for_status()
            results = response.json()
            total_records: int = int(results.get("total_count", 0))
            if total_records >= offset:
                more_records = True

            return more_records, results
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.Timeout,
            requests.exceptions.ChunkedEncodingError,
            requests.exceptions.ConnectTimeout,
        ) as exc:
            raise ConnectionError(exc) from exc

        except requests.exceptions.HTTPError as exc:
            raise HTTP500InternalServerError(exc) from exc

        except requests.exceptions.RequestException as exc:
            if "Please include a case-sensitive header of Authorization" in str(
                response.json()
            ):
                raise AuthenticationError(
                    "Please check the credentials used to make api call"
                ) from exc
            raise exc


class OneSignalHandler:
    """class to do the data pull"""

    def __init__(self, creds: dict, configs: dict):
        self.creds = creds
        self.session = requests.Session()
        self.configs = configs
        self.file_size_limit: int = self.configs.get("file_size_limit", 500000)

    def fetch_data(self) -> Generator[Dict[Any, Any], None, None]:
        """Method to fetch the data from source"""
        raise NotImplementedError


class ViewNotificationsHandler(OneSignalHandler):
    """Fetch View Notifications"""

    def __init__(self, creds: dict, configs: dict):
        self.offset: int = 0
        self.api_call_limit: int = configs.get("api_call_limit", 2000)
        super().__init__(creds=creds, configs=configs)

    def format_date(
        self, date_value: Union[str, None], result: dict
    ) -> Union[str, None]:
        """process date from int to YYYY-mm-dd"""
        if date_value is None:
            date_value = result.get("queued_at")
        if date_value is None:
            return None
        return datetime.fromtimestamp(int(date_value)).strftime("%Y%m%d")

    def check_limit_reached(
        self,
        partition_dataset: Dict[Any, Any],
        results: List[dict],
        startdate: int,
        enddate: int,
        more_records: bool = True,
    ) -> Tuple[bool, Dict]:
        """method to check if we have reached the end of api call

        Args:
            partition_dataset (Dict[Any, Any]): data_dictionary
            results (dict): results dictionary
            startdate (int): end date to when to stop api call
            enddate (int): end date to when to stop api call
            more_records (True): chec if we have reached end of api call

        Returns:
            Tuple[Dict, bool]: more records and the data dictionary
        """
        exit_condition: List[bool] = []
        for result in results:
            queued_at = result.get("queued_at")
            queued_at_date_str = self.format_date(queued_at, result)
            if queued_at_date_str is None or int(queued_at_date_str) > enddate:
                continue
            if int(queued_at_date_str) < startdate:
                exit_condition.append(True)
            else:
                result["completed_at_date"] = self.format_date(
                    result.get("completed_at"), result
                )
                result["queued_at_date"] = queued_at_date_str
                result["app_id"] = self.creds["app_id"]
                partition_dataset.setdefault(queued_at_date_str, []).append(result)

        if any(exit_condition) or self.api_call_limit <= 0:
            more_records = False

        return more_records, partition_dataset

    def fetch_data(self):
        """fecth data method"""
        # gather data per offset given the set of limits
        limit = self.configs.get("limit", 50)
        api_call_handler: APICallHandler = ViewNotificationsDownload(creds=self.creds)
        more_records: bool = True
        startdate: int = int(
            date_string_handler(self.configs.get("start_date", "2_days_ago")).strftime(
                "%Y%m%d"
            )
        )
        enddate: int = int(
            date_string_handler(self.configs.get("end_date", "today")).strftime(
                "%Y%m%d"
            )
        )
        print(f"working with startdate: {startdate} enddate {enddate}")
        partition_dataset: Dict[Any, Any] = {}
        while more_records:
            response = api_call_handler.api_call(
                session=self.session, limit=limit, offset=self.offset
            )
            if response is None:
                print("No data for view notifications")
            more_records, results = response[0], response[1]
            notifications = results.get("notifications")
            if notifications is not None and isinstance(notifications, list):
                more_records, partition_dataset = self.check_limit_reached(
                    partition_dataset, notifications, startdate, enddate, more_records
                )
                print(f"At offset {self.offset} of {results.get('total_count')}")

            self.offset += limit
            self.api_call_limit -= 1
            print(f"api call limit {self.api_call_limit}")

        if not partition_dataset:
            print("No view notifications data")
            return

        for date, date_dataset in partition_dataset.items():
            yield {"date": date, "data": date_dataset}


class CSVExportHandler(OneSignalHandler):
    """Class to make the CSV Export Job"""

    @staticmethod
    def add_created_at_date(row) -> Dict[Any, Any]:
        """function to add created_at_date"""
        try:
            row["created_at_date"] = str(
                datetime.strptime(row["created_at"], "%Y-%m-%d %H:%M:%S").date()
            ).replace("-", "")
        except ValueError:
            row["created_at_date"] = str(
                datetime.strptime(row["created_at"], "%Y-%m-%d %H:%M:%S.%f").date()
            ).replace("-", "")
        return row

    def fetch_data(self):
        """fetch data method for csv export"""
        bulk_creator: APICallHandler = CreateBulkDownload(creds=self.creds)
        api_call_handler: APICallHandler = FetchBulkDownload(creds=self.creds)
        bulk_job: Dict[str, str] = bulk_creator.api_call(self.session)
        if (bulk_job is None) or not isinstance(bulk_job, dict):
            raise CSVExportCreationError(
                f"Failed to create bulk job for {self.creds['app_id']}"
            )
        csv_file_url = bulk_job.get("csv_file_url")
        if csv_file_url is not None and isinstance(csv_file_url, str):
            print(f"created_job app_id: {self.creds['app_id']}: {bulk_job}")
            response = api_call_handler.api_call(self.session, url=csv_file_url)
            partition_dataset: Dict[Any, Any] = {}
            results = parse_zip_to_csv(response=response, file_type="gzip")
            if not results:
                print(f"No data for csv export for appi_id {self.creds['app_id']}")
                sys.exit(0)
            for result in results:
                result["app_id"] = self.creds["app_id"]  # inject app_id as metadata
                result = self.add_created_at_date(result)  # add created_at_date
                date_value = result.get("created_at_date")
                if date_value is not None:
                    partition_dataset.setdefault(date_value, []).append(result)

            for date, date_dataset in partition_dataset.items():
                yield {"date": date, "data": date_dataset}


class OneSignalHandlerFactory:
    """Factory to get us the Handler"""

    def get_endpoint_handler(self, creds: dict, configs: dict) -> OneSignalHandler:
        """Gets Us the Endpoint Hanlder to Use"""
        endpoint_handlers = {
            "csv_export": CSVExportHandler,
            "view_notifications": ViewNotificationsHandler,
        }

        return endpoint_handlers[configs["endpoint"]](creds=creds, configs=configs)



"""READER FILE FOR ZOHO RECRUIT"""
# pylint: disable=wrong-import-order,arguments-differ, broad-except, too-few-public-methods

from datetime import datetime
from typing import List, Dict, Union, Any, Generator

from sdc_dp_helpers.base_readers import APIRequestsSessionsReader
from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.date_managers import date_string_handler
from sdc_dp_helpers.zoho_recruit.zoho_recruit_sdk import ZohoRecruitAPI


class ZohoRecruitReader(APIRequestsSessionsReader):
    """Zoho Recruit Reader Class

    Args:
        BaseReader (_type_): API Reader Class
    """

    def __init__(self, secrets_filepath: str, config_filepath: str):
        self.configs: Dict[Any, Any] = load_file(config_filepath)
        self.secrets: Dict[Any, Any] = load_file(secrets_filepath)
        self.service = self._get_auth()
        self.success: List[bool] = []

    def _get_auth(self):
        handler_factory = ZohoRecruitAPI(session=self.sessions, secrets=self.secrets)
        return handler_factory.get_end_point_handler(
            end_point=self.configs["end_point"],
        )

    def _query_handler(
        self, configs: Dict[Any, Any]
    ) -> Generator[Dict[str, Union[str, List[Dict[Any, Any]]]], None, None]:
        """Method ot get the endpoint handler to use

        Args:
            config (Dict[Any, Any]): configs passed to the API endpoint handler

        Returns:
            List[Dict[Any, Any]]:  A list of dictionary of each response
        """
        results = self.service.fetch_data(configs=configs)
        for result in results:
            if result["data"]:
                yield {
                    "data": result["data"],
                    "module": result["module"],
                    "start_page": result["start_page"],
                    "end_point": self.configs["end_point"],
                    "date": datetime.now().date().strftime("%Y%m%d"),
                }
                self.is_success()
            else:
                self.not_success()

    def run_query(
        self,
    ) -> Generator[Dict[str, Union[str, List[Dict[Any, Any]]]], None, None]:
        """Method to run the query"""
        modified_since = self.configs.get("modified_since")
        if modified_since is None:
            modified_since = "50_years_ago"
        modified_since = date_string_handler(modified_since)
        modified_since = datetime.strftime(modified_since, "%Y-%m-%dT%H:%M:%S+00:00")

        self.configs["modified_since"] = modified_since
        results = self._query_handler(configs=self.configs)

        yield from results



"""MODULE TO GET DATA FROM BULK ENDPOINT"""
# pylint: disable=too-few-public-methods,arguments-differ,too-many-nested-blocks,inconsistent-return-statements,too-many-locals,too-many-branches
import os
import sys
from csv import DictReader
from io import TextIOWrapper
from zipfile import ZipFile
from pathlib import Path
from abc import ABC, abstractmethod
from typing import Union, List, Dict, Any
from zcrmsdk.src.com.zoho.crm.api.bulk_read import (
    BulkReadOperations,
    RequestWrapper,
    FileBodyWrapper,
    APIException,
    Query,
    CallBack,
    Criteria,
    ActionWrapper,
    SuccessResponse,
    ResponseWrapper,
)
from zcrmsdk.src.com.zoho.crm.api.exception.sdk_exception import SDKException
from zcrmsdk.src.com.zoho.crm.api.util.choice import Choice
from zcrmsdk.src.com.zoho.crm.api.users.user import User
from zcrmsdk.src.com.zoho.crm.api.layouts.layout import Layout
from zcrmsdk.src.com.zoho.crm.api.record.record import Record
from zcrmsdk.src.com.zoho.crm.api.roles.role import Role
from zcrmsdk.src.com.zoho.crm.api.profiles.profile import Profile

from sdc_dp_helpers.api_utilities.date_managers import date_string_handler
from sdc_dp_helpers.api_utilities.retry_managers import retry_handler, request_handler


class ResourceNotFoundError(Exception):
    """
    The requested resource doesn't exist.
    Resolution: The job ID specified is invalid. Specify a valid job ID.
    """


class InternalError(Exception):
    """
    Internal Server Error
    Resolution: Unexpected and unhandled exception in Server. Contact support team.
    """


class ErrorHandler:
    """class to handle errors"""

    def __init__(self):
        self.errors: list = []

    def append_error(self, module: str, error):
        """method to add errors to list"""
        self.errors.append({module: error})

    def get_errors(self):
        """method to return the error list"""
        return "\n".join([str(err) for err in self.errors])


class RequestHandler(ABC):
    """Base class to handle API calls"""

    @abstractmethod
    def api_call(self, **kwargs):
        """class method to make API calls"""
        raise NotImplementedError


class EndpointHandler:
    """class to create bulk job for export"""

    def make_api_call(self, **kwargs):
        """calls the request handler class"""
        raise NotImplementedError


class BulkEndpointHandler(EndpointHandler):
    """Endpoint handler for bulk api"""

    def __init__(self, configs: dict):
        self.configs = configs
        self.file_size_limit = configs.get("file_size_limit", 5242880)
        self.error_handler = ErrorHandler()
        self.create_bulk_read = CreateBulkReadRequestHandler(
            configs=configs, error_handler=self.error_handler
        )
        self.check_status = CheckBulkReadRequestHandler(
            error_handler=self.error_handler
        )
        self.download_result = DownloadResultRequestHandler(
            error_handler=self.error_handler
        )

    @staticmethod
    def process_more_columns(column_name: str, value):
        """Method to process the data"""
        if isinstance(value, (User, Layout, Profile, Role)):
            return column_name, {
                "id": value.get_id(),
                "name": value.get_name(),
            }
        if isinstance(value, Choice):
            return column_name, value.get_value()
        if isinstance(value, Record):
            return column_name, value.get_id()
        return column_name, str(value)

    def make_api_call(self, module_name: str, job_id: int, destination_folder: str):
        """make api call to fetch data"""
        yield from self.download_result.api_call(
            module_name=module_name,
            job_id=job_id,
            destination_folder=destination_folder,
        )

    def paginate(self, module_name: str):
        """paginate through the pages"""
        result_array: List[Dict[Any, Any]] = []
        current_size: int = 0
        more_records: bool = True
        page: int = self.configs.get("page", 1)
        page_limit: int = self.configs.get("page_limit")
        destination_folder: str = self.configs.get("destination_folder", "data_folder")
        while more_records:
            more_records = False
            job_creation = self.create_bulk_read.api_call(
                module_name=module_name, configs=self.configs, page=page
            )
            if (
                not isinstance(job_creation, dict)
                or job_creation.get("results", {"id": None}).get("id") is None
            ):
                break
            job, module_partition = (
                job_creation["results"],
                job_creation["module_partition"],
            )
            print(job)

            job_status: dict = self.check_status.api_call(
                module_name=module_name, job_id=job["id"]
            )
            if isinstance(job_status, dict) and job_status.get("download_url"):
                print(job_status)
                results = self.make_api_call(
                    module_name=module_name,
                    job_id=job["id"],
                    destination_folder=destination_folder,
                )
                counter = 0
                for result in results:
                    result_array.append(result)
                    current_size += sys.getsizeof(result)
                    if current_size >= self.file_size_limit:
                        yield {
                            "module_partition": module_partition,
                            "partition_counter": counter,
                            "data": result_array,
                        }
                        result_array: List[Dict[Any, Any]] = []
                        current_size = 0
                    counter += 1

                if result_array:
                    yield {
                        "module_partition": module_partition,
                        "partition_counter": counter,
                        "data": result_array,
                    }
                more_records = job_status.get("more_records", False)
                if more_records is False:
                    break

                if page_limit is not None and page >= page_limit:
                    print("early exit we have reached page limit")
                    break

                page += 1


class CreateBulkReadRequestHandler:
    """Handler to Create Bulk Job"""

    def __init__(self, configs: dict, error_handler):
        self.configs = configs
        self.error_handler = error_handler

    @staticmethod
    def process_more_columns(column_name: str, value):
        """Method to process the data"""
        if isinstance(value, (User, Layout, Profile, Role)):
            return column_name, {
                "id": value.get_id(),
                "name": value.get_name(),
            }
        if isinstance(value, Choice):
            return column_name, value.get_value()
        if isinstance(value, Record):
            return column_name, value.get_id()
        return column_name, str(value)

    def build_query(
        self, module_name: str, page: int, cvid: Union[str, None], configs: dict
    ):
        """method to build the query to be used for the api call"""
        # Get instance of CallBack Class
        call_back = CallBack()
        call_back.set_url("https://www.example.com/callback")  # Set valid callback URL
        call_back.set_method(Choice("post"))
        request = RequestWrapper()
        # The Bulk Read Job's details is posted to this URL
        # on successful completion / failure of the job.
        request.set_callback(call_back)
        query = Query()  # Get instance of Query Class
        query.set_module(module_name)

        # Specifies the unique ID of the custom view, whose records you want to export.
        if cvid is not None:
            query.set_cvid(str(cvid))

        # Specifies the API Name of the fields to be fetched
        if configs.get("select_fields"):
            query.set_fields(configs.get("select_fields"))
        # To set page value, By default value is 1.
        query.set_page(page)

        # Get instance of Criteria Class
        criteria = Criteria()
        criteria_group = configs["group"]
        criteria_datetime_fields = configs["criteria_datetime_fields"]
        for group in criteria_group:
            value = group["value"]
            api_name = group["api_name"]
            if api_name in criteria_datetime_fields:
                time_format = criteria_datetime_fields[api_name]
                value = [
                    date_string_handler(value[0], time_format=time_format).strftime(
                        time_format
                    ),
                    date_string_handler(value[1], time_format=time_format).strftime(
                        time_format
                    ),
                ]
                print(value)
            criteria.set_api_name(api_name)
            criteria.set_comparator(Choice(group["comparator"]))
            criteria.set_value(value)

        query.set_criteria(criteria)
        request.set_query(query)
        request.set_file_type(Choice("csv"))

        return request

    @retry_handler(
        exceptions=ConnectionError, total_tries=3, initial_wait=2, backoff_factor=2
    )
    @retry_handler(
        exceptions=InternalError, total_tries=3, initial_wait=2, backoff_factor=2
    )
    @request_handler(
        wait=int(os.environ.get("API_WAIT_TIME", 5)),
        backoff_factor=0.01,
        backoff_method="random",
    )
    def api_call(self, module_name: str, configs: dict, page=1):
        """
        This method is used to create a bulk read job to export records.
        :param module_api_name: The API Name of the record's module
        module_name = 'Leads'
        """
        try:
            cvid = None
            module_prefix = ""
            modulename_combination = module_name.rsplit("-", maxsplit=1)
            module_name = modulename_combination[0]
            if len(modulename_combination) > 1:
                module_prefix = modulename_combination[1]
            if configs.get("cvid") and module_prefix.startswith("CustomView"):
                cvid = configs["cvid"].get(f"{module_name}-{module_prefix}")

            request = self.build_query(
                module_name=module_name, page=page, cvid=cvid, configs=configs
            )
            bulk_read_operations = BulkReadOperations()
            response = bulk_read_operations.create_bulk_read_job(request)

            if response is not None:
                response_object = response.get_object()

                if response_object is not None:
                    # Check if expected ActionWrapper instance is received.
                    if isinstance(response_object, ActionWrapper):
                        action_response_list = response_object.get_data()

                        for action_response in action_response_list:
                            # Check if the request is successful
                            if isinstance(action_response, SuccessResponse):
                                details = action_response.get_details()
                                result = {
                                    "status": action_response.get_status().get_value(),
                                    "code": action_response.get_code().get_value(),
                                    "id": int(details["id"]),
                                    "details": dict(
                                        [
                                            self.process_more_columns(key, value)
                                            for key, value in details.items()
                                        ]
                                    ),
                                    "message": action_response.get_message().get_value(),
                                }
                                module_partition = f"{module_name}"
                                if cvid is not None:
                                    module_partition = f"{module_name}_{cvid}"
                                return {
                                    "results": result,
                                    "module_partition": module_partition,
                                }

                            if isinstance(action_response, APIException):
                                error_code = action_response.get_code().get_value()
                                error_details = action_response.get_details()
                                # error_status = action_response.get_status().get_value()
                                self.error_handler.append_error(
                                    module_name, f"{error_code}, {str(error_details)}"
                                )

                    elif isinstance(response_object, APIException):
                        error_code = response_object.get_code().get_value()
                        error_details = response_object.get_details()
                        self.error_handler.append_error(
                            module_name, f"{error_code}, {str(error_details)}"
                        )
                        if error_code == "INTERNAL_ERROR":
                            raise InternalError(
                                f"{module_name} {error_code}, {str(error_details)}"
                            )
        except SDKException as exc:
            if "ConnectionError" in str(exc):
                raise ConnectionError(exc) from exc


class CheckBulkReadRequestHandler:
    """Check Bulk Read Job"""

    def __init__(self, error_handler):
        self.error_handler = error_handler

    @retry_handler(
        exceptions=ConnectionError, total_tries=3, initial_wait=2, backoff_factor=2
    )
    @retry_handler(
        exceptions=InternalError, total_tries=3, initial_wait=2, backoff_factor=2
    )
    @retry_handler(
        exceptions=ResourceNotFoundError,
        total_tries=10,
        initial_wait=2,
        backoff_factor=2,
    )
    @request_handler(
        wait=int(os.environ.get("API_WAIT_TIME", 5)),
        backoff_factor=0.01,
        backoff_method="random",
    )
    def api_call(self, module_name: str, job_id: int):
        """
        This method is used to get the details of a bulk read job performed previously.
        :param job_id: The unique ID of the bulk read job.
        job_id = 3409643000002461001
        """
        try:
            results = None
            bulk_read_operations = BulkReadOperations()
            response = bulk_read_operations.get_bulk_read_job_details(job_id)

            if response is not None:
                if response.get_status_code() in [204, 304]:
                    print(
                        "No Content"
                        if response.get_status_code() == 204
                        else "Not Modified"
                    )
                    return results

                response_object = response.get_object()

                if response_object is not None:
                    # Check if expected ResponseWrapper instance is received
                    if isinstance(response_object, ResponseWrapper):
                        # Get the list of JobDetail instances
                        job_details_list = response_object.get_data()

                        for job_detail in job_details_list:
                            # Get the Job ID of each jobDetail
                            criteria = job_detail.get_query().get_criteria()
                            results = {
                                "id": job_detail.get_id(),
                                "read_operation": job_detail.get_operation(),
                                "read_state": job_detail.get_state().get_value(),
                                "query": {
                                    "criteria": {
                                        "value": criteria.get_value(),
                                        "api_name": criteria.get_api_name(),
                                    },
                                    "cvid": job_detail.get_query().get_cvid(),
                                },
                            }

                            result = job_detail.get_result()

                            if result is not None:
                                results.update(
                                    {
                                        "page": str(result.get_page()),
                                        "count": str(result.get_count()),
                                        "download_url": result.get_download_url(),
                                        "per_page": str(result.get_per_page()),
                                        "more_records": result.get_more_records(),
                                    }
                                )

                            if (
                                isinstance(results["id"], int)
                                and results["read_state"] != "COMPLETED"
                            ):
                                raise ResourceNotFoundError(f"{module_name}, {results}")
                        return results

                    if isinstance(response_object, APIException):
                        error_code = response_object.get_code().get_value()
                        error_details = response_object.get_details()
                        self.error_handler.append_error(
                            module_name, f"{job_id} {error_code}, {str(error_details)}"
                        )
                        if error_code == "RESOURCE_NOT_FOUND":
                            raise ResourceNotFoundError(
                                f"{module_name} {error_code}, {str(error_details)}"
                            )
                        if error_code == "INTERNAL_ERROR":
                            raise InternalError(
                                f"{module_name} {error_code}, {str(error_details)}"
                            )

        except SDKException as exc:
            if "ConnectionError" in str(exc):
                raise ConnectionError(exc) from exc


class DownloadResultRequestHandler:
    """Download result handler for bulk api"""

    def __init__(self, error_handler):
        self.error_handler = error_handler

    @retry_handler(
        exceptions=ConnectionError, total_tries=3, initial_wait=2, backoff_factor=2
    )
    @retry_handler(
        exceptions=InternalError, total_tries=3, initial_wait=2, backoff_factor=2
    )
    @retry_handler(
        exceptions=ResourceNotFoundError,
        total_tries=10,
        initial_wait=2,
        backoff_factor=2,
    )
    @request_handler(
        wait=int(os.environ.get("API_WAIT_TIME", 5)),
        backoff_factor=0.01,
        backoff_method="random",
    )
    def api_call(self, module_name: str, job_id: int, destination_folder: str):
        """
        This method is used to download the result of Bulk Read operation
        :param job_id: The unique ID of the bulk read job.
        :param destination_folder: The absolute path where downloaded file has to be stored.
        job_id = 3409643000002461001
        """
        try:
            # Get instance of BulkReadOperations Class
            bulk_read_operations = BulkReadOperations()
            response = bulk_read_operations.download_result(job_id)

            if response is not None:
                if response.get_status_code() in [204, 304]:
                    print(
                        "No Content"
                        if response.get_status_code() == 204
                        else "Not Modified"
                    )
                    return

                # Get object from response
                response_object = response.get_object()

                if response_object is not None:
                    # Check if expected FileBodyWrapper instance is received.
                    if isinstance(response_object, FileBodyWrapper):
                        stream_wrapper = response_object.get_file()
                        # Construct the file name by joining the destinationFolder
                        # and the name from StreamWrapper instance
                        Path(f"{destination_folder}").mkdir(parents=True, exist_ok=True)
                        file_name = os.path.join(
                            destination_folder, stream_wrapper.get_name()
                        )

                        with open(file_name, "wb") as dest:
                            # Get the stream from StreamWrapper instance
                            for chunk in stream_wrapper.get_stream():
                                dest.write(chunk)
                            dest.close()

                        with ZipFile(file_name) as zipfile:
                            for file in zipfile.filelist:
                                with zipfile.open(file) as file_contents:
                                    dict_reader = DictReader(
                                        TextIOWrapper(file_contents, encoding="utf-8")
                                    )
                                    yield from dict_reader

                    elif isinstance(response_object, APIException):
                        error_code = response_object.get_code().get_value()
                        error_details = response_object.get_details()
                        self.error_handler.append_error(
                            module_name, f"{job_id} {error_code}, {str(error_details)}"
                        )
                        if error_code == "RESOURCE_NOT_FOUND":
                            raise ResourceNotFoundError(
                                f"{module_name} {error_code}, {str(error_details)}"
                            )
                        if error_code == "INTERNAL_ERROR":
                            raise InternalError(
                                f"{module_name} {error_code}, {str(error_details)}"
                            )
        except SDKException as exc:
            if "ConnectionError" in str(exc):
                raise ConnectionError(exc) from exc



# pylint: disable=too-few-public-methods,arguments-differ,import-error,too-many-arguments,too-many-locals,line-too-long, disable=useless-elif-before-else, disable=R1705, disable=R1720, disable=W1309, disable=C0103
""" Falcon Reader SDK"""
from abc import abstractmethod
from typing import List, Dict
import os
from datetime import datetime
from random import randint, random
import requests
from sdc_dp_helpers.api_utilities.retry_managers import request_handler, retry_handler


class FalconResponseInProgress(Exception):
    """Class for Falcon API call in progress """


class FalconTimeoutException(Exception):
    """Class for Ficlon timedout Exception """


class FalconQuotaException(Exception):
    """"Class for Ficlon Quota Exception """


class RequestHandler:
    """Interface for  API Call method"""
    api_calls = 0

    def __init__(self, session, creds):
        self.session: requests.Session = session
        self.base_url: str = "https://api.falcon.io/"
        self._creds = creds

    @abstractmethod
    def make_api_call(self, **kwargs):
        """Make API Call"""
        raise NotImplementedError


class FalconAPICall:
    """Class for Making API Call"""

    def __init__(self, session, creds, config):
        self.session = session
        self._creds = creds
        self.config = config
        self.channel_ids = self.get_channel_ids()
        self.start = 0
        self.curr_channels = []

    def get_channel_ids(self):
        """This gets channel ids"""
        channel_id_methods = {"v1": ChannelIdsV1, "v2": ChannelIdsV2}
        version = self.config["version"]
        if version not in channel_id_methods:
            raise KeyError(
                f"unkown value for 'version' in configs, expecting 'v1' or 'v2'. Got {version}"
            )
        channel_ids_caller = channel_id_methods[version](
            session=self.session, creds=self._creds
        )
        channel_ids = channel_ids_caller.make_api_call()

        return channel_ids

    def get_id_lookup(self):
        RequestHandler.api_calls += 1
        print("get_id_lookup")
        endpoint_url = f"channels?apikey={self._creds['api_key']}"
        url = f"https://api.falcon.io/{endpoint_url}"
        response = self.session.get(url=url, params={"limit": "9999"})

        items = []
        if int(response.status_code) == 200:
            response_data = response.json()
            for item in response_data.get("items", []):
                items.append(
                    {
                        'channelid_v1': item['id'],
                        'channelid_v2': item['uuid'],
                        'network': item['network'],
                        'brand': item['name']

                    }
                )
            return items

        if int(response.status_code) == 429:
            if 'Retry-After' in response.headers:
                wait_time = response.headers.get(
                    'Retry-After', randint(4, 29)) * (1 + randint(2, 8) * random())
                os.environ['API_WAIT_TIME'] = str(wait_time)
                raise FalconTimeoutException(
                    f"Rolling Window Quota [429] reached. Waiting {wait_time} seconds. ")
            else:
                raise FalconQuotaException(
                    f"Falcon Quota Reached when getting channel id v2. "
                    f"Status code: {response.status_code}, Reason: {response.reason}. "
                )
        raise ConnectionError(
            f"Falcon API failed to return content id by channel ids. "
            f"Status code: {response.status_code}, Reason: {response.reason}. "
        )

    def get_data(
        self, start_date: str, end_date: str, network: str, channel_ids: dict
    ) -> dict:
        """Gets the API Call Handler to Use"""
        endpoint_handlers = {
            "channel_insights": ChannelInsights,
            "content_insights": ContentInsights,
            "published_posts": PublishedPosts,
        }
        endpoint_name = self.config["endpoint_name"]
        if endpoint_name not in endpoint_handlers:
            raise KeyError(
                f"""
                unkown endpoint name in configs, expecting 'channel_insights',
                'content_insights' or 'published_posts'. Got {endpoint_name}
                """
            )
        endpoint_caller = endpoint_handlers[endpoint_name](
            session=self.session, creds=self._creds, config=self.config
        )
        data = endpoint_caller.get_dataset(
            start_date, end_date, network, channel_ids)

        return data


class ChannelIdsV1(RequestHandler):
    """Class for v1 channel ids"""
    @retry_handler(exceptions=FalconQuotaException, total_tries=1, should_raise=True)
    @retry_handler(exceptions=FalconTimeoutException, total_tries=5)
    @retry_handler(exceptions=ConnectionError, total_tries=5)
    def make_api_call(self, **kwargs):
        """
        Gather all available channel ids v1.
        """
        RequestHandler.api_calls += 1
        print("GET v1: channel ids. ")
        endpoint_url = f"channels?apikey={self._creds['api_key']}"
        url = f"{self.base_url}{endpoint_url}"
        response = self.session.get(url=url, params={"limit": "9999"})

        if int(response.status_code) == 200:
            response_data = response.json()
            channel_ids = set()
            for item in response_data.get("items", []):
                channel_ids.add(item["id"])

            return list(channel_ids)

        if int(response.status_code) == 429:
            if 'Retry-After' in response.headers:
                wait_time = response.headers.get(
                    'Retry-After', randint(4, 29)) * (1 + randint(2, 8) * random())
                os.environ['API_WAIT_TIME'] = str(wait_time)
                raise FalconTimeoutException(
                    f"Rolling Window Quota [429] reached. Waiting {wait_time} seconds. ")
            else:
                raise FalconQuotaException(
                    f"Falcon Quota Reached when getting channel id v1. "
                    f"Status code: {response.status_code}, Reason: {response.reason}. "
                )
        raise ConnectionError(
            f"Falcon API failed to return content id by channel ids. "
            f"Status code: {response.status_code}, Reason: {response.reason}. "
        )


class ChannelIdsV2(RequestHandler):
    """Class for v2 channel ids"""

    @retry_handler(exceptions=FalconQuotaException, total_tries=1, should_raise=True)
    @retry_handler(exceptions=FalconTimeoutException, total_tries=5)
    @retry_handler(exceptions=ConnectionError, total_tries=5)
    def make_api_call(self, **kwargs):
        """
        Gather all available channel ids v2.
        """
        RequestHandler.api_calls += 1
        print("GET v2: channel ids. ")
        endpoint_url = f"channels?apikey={self._creds['api_key']}"
        url = f"{self.base_url}{endpoint_url}"
        response = self.session.get(url=url, params={"limit": "9999"})

        if int(response.status_code) == 200:
            response_data = response.json()
            channel_ids = {}
            for item in response_data.get("items", []):
                channel_ids[item["uuid"]] = item["name"]
            return channel_ids

        if int(response.status_code) == 429:
            if 'Retry-After' in response.headers:
                wait_time = response.headers.get(
                    'Retry-After', randint(4, 29)) * (1 + randint(2, 8) * random())
                os.environ['API_WAIT_TIME'] = str(wait_time)
                raise FalconTimeoutException(
                    f"Rolling Window Quota [429] reached. Waiting {wait_time} seconds. ")
            else:
                raise FalconQuotaException(
                    f"Falcon Quota Reached when getting channel id v2. "
                    f"Status code: {response.status_code}, Reason: {response.reason}. "
                )
        raise ConnectionError(
            f"Falcon API failed to return content id by channel ids. "
            f"Status code: {response.status_code}, Reason: {response.reason}. "
        )


class ContentIdByChannelId(RequestHandler):
    """Class to get Content Ids by Channel id"""

    @retry_handler(exceptions=FalconQuotaException, total_tries=1, should_raise=True)
    @retry_handler(exceptions=FalconTimeoutException, total_tries=5)
    @retry_handler(exceptions=ConnectionError, total_tries=5, initial_wait=5)
    def make_api_call(
        self, start_date: str, end_date: str, network: str, channel_ids: Dict[str, str]
    ) -> dict:
        print("GET: content ids by channel id. ")
        date_filters = f"since={start_date}&until={end_date}"
        endpoint_url = (
            f"publish/items?apikey={self._creds['api_key']}"
            f"&statuses=published"
            f"&networks={network}"
            f"&{date_filters}"
        )
        content_ids_by_channel_id = {}
        while endpoint_url is not None:
            RequestHandler.api_calls += 1
            url = f"{self.base_url}{endpoint_url}"
            response = self.session.get(url=url)
            if int(response.status_code) == 200:
                response_data = response.json()
                for item in response_data.get("items"):
                    content_id = item.get("id")
                    channel_id = item.get("channels")
                    if channel_id is not None:
                        channel_id = channel_id[0]
                        if network == "linkedIn":
                            channel_id = channel_id.replace("-", "")
                        if channel_id in channel_ids:
                            content_ids_by_channel_id.setdefault(channel_id, []).append(
                                content_id
                            )
                endpoint_url = response_data.get(
                    "next", {"href": None}).get("href")
            elif int(response.status_code) == 429:
                if 'Retry-After' in response.headers:
                    wait_time = response.headers.get(
                        'Retry-After', randint(4, 29)) * (1 + randint(2, 8) * random())
                    os.environ['API_WAIT_TIME'] = str(wait_time)
                    raise FalconTimeoutException(
                        f"Rolling Window Quota [429] reached. Waiting {wait_time} seconds. ")
                else:
                    raise FalconQuotaException(
                        f"Falcon Quota Reached when getting content id by channel id. "
                        f"Status code: {response.status_code}, Reason: {response.reason}. "
                    )
            else:
                raise ConnectionError(
                    f"Falcon API failed to get content id by channel id. "
                    f"Status code: {response.status_code}, Reason: {response.reason}. "
                )
        return content_ids_by_channel_id


class PublishedPostsByChannelId(RequestHandler):
    """Gets the published posts by channel id"""

    @request_handler(
        wait=int(os.environ.get("REQUEST_WAIT_TIME", 2)),
        backoff_factor=float(os.environ.get("REQUEST_BACKOFF_FACTOR", 0.01)),
        backoff_method=os.environ.get("REQUEST_BACKOFF_METHOD", 0.01),
    )
    @retry_handler(exceptions=FalconQuotaException, total_tries=1, should_raise=True)
    @retry_handler(exceptions=FalconTimeoutException, total_tries=5)
    @retry_handler(exceptions=ConnectionError, total_tries=5, initial_wait=900)
    def make_api_call(
        self,
        start_date: str,
        end_date: str,
        network: str,
        channel_ids: list,
        statuses: str,
        limit: int,
        index: int,
    ) -> list:
        """Gets the published posts by channel id
        SEE: https://falconio.docs.apiary.io/
        #reference/content-api/get-content-metrics-by-channel-ids
        :param: channel_id - integer
        :param: date - string '2022-07-20' but converted to isoformat '2022-07-20T00:00:00'
        :returns: list of dictionaries
        """

        dataset: list = []
        endpoint_url: str = f"publish/items?apikey={self._creds['api_key']}"
        while endpoint_url:
            RequestHandler.api_calls += 1
            print(
                f"INFO: channel id index: {index}, "
                f"channel id: {channel_ids},  date: {start_date}, offset: {len(dataset)}. "
            )
            response = self.session.get(
                url=f"{self.base_url}{endpoint_url}",
                headers={"Content-Type": "application/json"},
                params={
                    "channels": channel_ids,
                    "since": start_date,
                    "until": end_date,
                    "networks": network,
                    "statuses": statuses,
                    "limit": limit,
                },
            )
            status_code: int = response.status_code
            reason: str = response.reason
            if int(response.status_code) == 200:
                results: dict = response.json()
                items_data: list = results.get("items", [])
                dataset.extend(items_data)

                endpoint_url = results.get("next", {"href": None}).get("href")

                if len(items_data) == 0 or endpoint_url is None:
                    break

            elif reason == "Not Found" and status_code == 404:
                print(f"No data for channel id: {channel_ids}, skipping.\n")
                break
            elif int(response.status_code) == 429:
                if 'Retry-After' in response.headers:
                    wait_time = response.headers.get(
                        'Retry-After', randint(4, 29)) * (1 + randint(2, 8) * random())
                    os.environ['API_WAIT_TIME'] = str(wait_time)
                    raise FalconTimeoutException(
                        f"Rolling Window Quota [429] reached. Waiting {wait_time} seconds. ")
                else:
                    raise FalconQuotaException(
                        f"Falcon Quota Reached when getting published posts by channel id. "
                        f"Status code: {response.status_code}, Reason: {response.reason}. "
                    )
            else:
                raise ConnectionError(
                    f"Falcon API failed to get publish post by channel id. "
                    f"Status code: {response.status_code}, Reason: {response.reason}. "
                )
        return dataset


class ContentInsightsRequestId(RequestHandler):
    """This class gets Content Insights Request Id"""

    @retry_handler(exceptions=FalconQuotaException, total_tries=1, should_raise=True)
    @retry_handler(exceptions=FalconTimeoutException, total_tries=5)
    @retry_handler(exceptions=ConnectionError, total_tries=5, initial_wait=5)
    def make_api_call(
        self,
        start_date: str,
        end_date: str,
        metric_ids: List[str],
        content_ids_by_channel_id: Dict[str, List],
    ) -> Dict[str, str]:
        RequestHandler.api_calls += 1
        print("GET: insights request id. ")
        endpoint_url = f"measure/v2/insights/content?apikey={self._creds['api_key']}"
        url = f"{self.base_url}{endpoint_url}"
        body = {
            "since": start_date,
            "until": end_date,
            "metricIds": metric_ids,
            "channels": [],
        }
        for key, value in content_ids_by_channel_id.items():
            body["channels"].append({"id": key, "contentIds": value})
        response = self.session.post(
            url=url, headers={"Content-Type": "application/json"}, json=body
        )
        # Check if we hit the rate limit

        if int(response.status_code) == 200:
            response_data = response.json()
            insights_request_id = response_data["insightsRequestId"]
            return insights_request_id
            # Check if we hit the rate limit
        elif int(response.status_code) == 429:
            if 'Retry-After' in response.headers:
                wait_time = response.headers.get(
                    'Retry-After', randint(4, 29)) * (1 + randint(2, 8) * random())
                os.environ['API_WAIT_TIME'] = str(wait_time)
                raise FalconTimeoutException(
                    f"Rolling Window Quota [429] reached. Waiting {wait_time} seconds. ")
            else:
                raise FalconQuotaException(
                    f"Falcon Quota Reached when getting content insight request ids. "
                    f"Status code: {response.status_code}, Reason: {response.reason}. "
                )
        else:
            raise ConnectionError(
                f"Falcon API failed to return content insight request ids. "
                f"Status code: {response.status_code}, Reason: {response.reason}. "
            )


class ChannelInsightsRequestId(RequestHandler):
    """This class gets Channel Insights Request Id"""

    @retry_handler(exceptions=FalconQuotaException, total_tries=1, should_raise=True)
    @retry_handler(exceptions=FalconTimeoutException, total_tries=5)
    @retry_handler(exceptions=ConnectionError, total_tries=5, initial_wait=5)
    def make_api_call(
        self,
        start_date: str,
        end_date: str,
        metric_ids: List[str],
        channel_ids: List[str],
    ) -> Dict[str, str]:
        RequestHandler.api_calls += 1
        print("GET: insights request id. ")
        start_date = datetime.strftime(
            datetime.strptime(start_date, "%Y-%m-%dT%H:%M:%SZ"), "%Y-%m-%d"
        )
        end_date = start_date
        endpoint_url = f"measure/v2/insights/channel?apikey={self._creds['api_key']}"
        url = f"{self.base_url}{endpoint_url}"
        body = {
            "since": start_date,
            "until": end_date,
            "metricIds": metric_ids,
            "channelIds": channel_ids,
        }
        response = self.session.post(
            url=url, headers={"Content-Type": "application/json"}, json=body
        )

        # Check if we hit the rate limit

        if int(response.status_code) == 200:
            response_data = response.json()
            insights_request_id = response_data["insightsRequestId"]
            return insights_request_id
        if int(response.status_code) == 429:
            if 'Retry-After' in response.headers:
                wait_time = response.headers.get(
                    'Retry-After', randint(4, 29)) * (1 + randint(2, 8) * random())
                os.environ['API_WAIT_TIME'] = str(wait_time)
                raise FalconTimeoutException(
                    f"Rolling Window Quota [429] reached. Waiting {wait_time} seconds. ")
            else:
                raise FalconQuotaException(
                    f"Falcon Quota Reached. "
                )
        raise ConnectionError(
            f"Falcon API failed to return channel insight request ids. "
            f"Status code: {response.status_code}, Reason: {response.reason}. "
        )


class Insights(RequestHandler):
    """This class gets insights data"""

    @retry_handler(exceptions=FalconResponseInProgress, total_tries=5, initial_wait=5)
    @retry_handler(exceptions=FalconQuotaException, total_tries=1, should_raise=True)
    @retry_handler(exceptions=FalconTimeoutException, total_tries=5)
    @retry_handler(exceptions=ConnectionError, total_tries=5, initial_wait=5)
    def make_api_call(self, insights_request_id: Dict[str, str] = None):
        RequestHandler.api_calls += 1
        print("GET: insights. ")
        endpoint_url = (
            f"measure/v2/insights/{insights_request_id}?apikey={self._creds['api_key']}"
        )
        url = f"{self.base_url}{endpoint_url}"
        response = self.session.get(url=url)
        response_data = response.json()
        if int(response.status_code) == 200:
            response_data = response.json()
            if response_data["status"] == "READY":
                return response_data
            elif response_data["status"] == "IN_PROGRESS":
                raise FalconResponseInProgress(
                    f"Response for Insights is {response_data['status']}. "
                )
            else:
                # Data is not ready.
                raise ConnectionError(
                    f"Falcon API failed to return content ids. "
                    f"Status code: {response.status_code}, Reason: Response status is {response_data['status']}. "
                )
        # Check if we hit the rate limit
        elif int(response.status_code) == 429:
            if 'Retry-After' in response.headers:
                wait_time = response.headers.get(
                    'Retry-After', randint(4, 29)) * (1 + randint(2, 8) * random())
                os.environ['API_WAIT_TIME'] = str(wait_time)
                raise FalconTimeoutException(
                    f"Rolling Window Quota [429] reached. Waiting {wait_time} seconds. ")
            else:
                raise FalconQuotaException(
                    f"Falcon Quota Reached when calling Insights. "
                    f"Status code: {response.status_code}, Reason: {response.reason}. "
                )
        else:
            raise ConnectionError(
                f"Falcon API failed to return insights. "
                f"Status code: {response.status_code}, Reason: {response.reason}. "
            )


class ContentInsights(FalconAPICall):
    """This class  gets content insights and normalises the data"""

    @staticmethod
    def _normalize_data(dataset: dict, network: str, channel_ids: dict) -> List[Dict]:
        temp = {}

        for metric, data_items in dataset.items():
            if not data_items:
                # if we have nothing in value
                continue
            for data in data_items:
                data[metric] = data.pop("value")
                data["network"] = network
                data["brand"] = channel_ids.get(data["channelId"], None)
                check_point = tuple(
                    map(data.get, ["channelId", "date", "contentId"]))
                if not temp.get(check_point):
                    temp[check_point] = data
                else:
                    missing_keys = set(data).difference(set(temp[check_point]))
                    # flatten
                    for key in missing_keys:
                        temp[check_point].update({key: data[key]})
        normalized_data = list(temp.values())
        return normalized_data

    def insights_handler(
        self,
        **kwargs,
    ):
        insights_requests_handler = ContentInsightsRequestId(
            creds=self._creds, session=self.session
        )
        insights_request_id = insights_requests_handler.make_api_call(**kwargs)
        results = Insights(
            creds=self._creds, session=self.session
        ).make_api_call(insights_request_id)
        if results is None or results.get("data") is None:
            insights_request_id = insights_requests_handler.make_api_call(
                **kwargs)
            results = Insights(
                creds=self._creds, session=self.session
            ).make_api_call(insights_request_id)

        return results

    def get_dataset(
        self, start_date: str, end_date: str, network: str, channel_ids: dict
    ) -> dict:
        """Fucntion handles content insights query"""
        channel_steps = self.config.get("channel_steps", 10)
        content_steps = self.config.get("content_steps", 250)
        metric_ids = self.config["metrics"]
        content_ids_by_channel_id = ContentIdByChannelId(
            creds=self._creds, session=self.session
        ).make_api_call(
            start_date=start_date,
            end_date=end_date,
            network=network,
            channel_ids=channel_ids,
        )
        if content_ids_by_channel_id is None:
            raise ConnectionError("Failed to get content id by channel id")
        channels = list(content_ids_by_channel_id.keys())
        channel_id_length = len(channels)
        dataset = []
        data = []
        for _ in range(0, channel_id_length, channel_steps):
            self.curr_channels = channels[self.start: self.start + channel_steps]
            request_content_ids_by_channel_id = {
                key: value
                for key, value in content_ids_by_channel_id.items()
                if key in self.curr_channels
            }

            for key, value in request_content_ids_by_channel_id.items():
                content_ids_length = len(value)
                start = 0
                curr_content_ids = []
                curr_request_content_ids_by_channel_id = {}
                for _ in range(0, content_ids_length, content_steps):
                    curr_content_ids = value[start: start + content_steps]
                    curr_request_content_ids_by_channel_id[key] = curr_content_ids
                    if len(curr_request_content_ids_by_channel_id) < 1:
                        continue
                    results = self.insights_handler(
                        start_date=start_date,
                        end_date=end_date,
                        metric_ids=metric_ids,
                        content_ids_by_channel_id=curr_request_content_ids_by_channel_id,
                    )
                    if results is not None:
                        data = results["data"].get("insights")
                        if data is not None:
                            dataset.extend(self._normalize_data(
                                data, network, channel_ids))
                start = start + content_steps
            self.start = self.start + channel_steps
        return dataset


class ChannelInsights(FalconAPICall):
    """This class  gets channel insights and normalises the data"""

    @staticmethod
    def _normalize_data(dataset: dict, network: str, channel_ids: dict) -> List[Dict]:
        templist = []

        for metric, data_items in dataset.items():
            if not data_items:
                # if we have nothing in value
                continue
            for data in data_items:
                data[metric] = data.pop("value")
                data["network"] = network
                data["brand"] = channel_ids.get(data["channelId"], None)
                templist.append(data)
        normalized_data = templist

        return normalized_data

    def get_dataset(
        self, start_date: str, end_date: str, network: str, channel_ids: dict
    ) -> dict:
        """Fucntion handles channel insights query"""
        channels = list(channel_ids.keys())
        channel_id_length = len(channels)
        channel_steps = self.config.get("channel_steps", 10)
        metric_ids = self.config["metrics"]
        dataset = []
        data = []
        for _ in range(0, channel_id_length, channel_steps):
            self.curr_channels = channels[self.start: self.start + channel_steps]
            insights_request_id = ChannelInsightsRequestId(
                creds=self._creds, session=self.session
            ).make_api_call(
                start_date=start_date,
                end_date=end_date,
                metric_ids=metric_ids,
                channel_ids=self.curr_channels,
            )
            results = Insights(creds=self._creds, session=self.session).make_api_call(
                insights_request_id
            )
            if results is None:
                insights_request_id = ChannelInsightsRequestId(
                    creds=self._creds, session=self.session
                ).make_api_call(
                    start_date=start_date,
                    end_date=end_date,
                    metric_ids=metric_ids,
                    channel_ids=self.curr_channels,
                )
                results = Insights(
                    creds=self._creds, session=self.session
                ).make_api_call(insights_request_id)
            if results is None:
                raise ConnectionError(
                    f"Falcon API failed to return channel insights. "
                )
            data = results["data"]["insights"]
            dataset.extend(self._normalize_data(data, network, channel_ids))
            self.start = self.start + channel_steps
        return dataset


class PublishedPosts(FalconAPICall):
    """This class gets published posts data"""

    def get_dataset(
        self, start_date: str, end_date: str, network: str, channel_ids: list
    ) -> dict:
        """Fucntion handles published posts query"""
        dataset = []
        if channel_ids is not None:
            print(
                f"Attempting to gather metrics from {len(channel_ids)} channel ids. ")
            statuses = self.config.get("statuses", "published")
            limit = self.config.get("limit", 2000)
            try:
                for channel_id in channel_ids:
                    self.curr_channels = [channel_id]
                    results = PublishedPostsByChannelId(
                        creds=self._creds, session=self.session
                    ).make_api_call(
                        start_date=start_date,
                        end_date=end_date,
                        network=network,
                        channel_ids=[channel_id],
                        statuses=statuses,
                        limit=limit,
                        index=self.start,
                    )
                    if results:
                        print(len(results))
                        dataset.extend(results)
                    self.start += 1
            except FalconQuotaException as e:
                raise e
        return dataset



# pylint: disable=import-error, too-few-public-methods, unused-import, too-many-arguments, line-too-long
"""
    CUSTOM WRITER CLASSES
        - Class which manages writer tasks like
        auth, write metadata, write file, create dir structure
"""
import re
from typing import Union, List, Any, Dict, Tuple, Optional
from sdc_dp_helpers.base_writers import BaseWriter


class GoogleKnowledgeGraphWriter(BaseWriter):
    """v1  GoogleKnowledgeGraphWriter Class"""
    def __init__(
        self,
        bucket: str,
        folder_path: str,
        destination: str,
        configs: Optional[Dict[str, Any]] = None,
        clear_destination: bool = False,
    ):
        self.success: List[bool] = []
        self.current_destination: Union[str, None] = None
        super().__init__(
            bucket=bucket,
            folder_path=folder_path,
            destination=destination,
            configs=configs,
            clear_destination=clear_destination,
        )
    def verify_data(
        self, payload: Dict[Any, Any]
    ) -> Tuple[str, Union[List[Dict[Any, Any]], Dict[Any, Any], Any]]:
        """verifies data before writting to destination also designs the destination file path

        Args:
            payload (Dict[Any, Any]):

        Raises:
            KeyError: if we do not find the exact keys we expect from the payload
            TypeError: if provided data object is not a list

        Returns:
            Tuple[str, Union[List[Dict[Any, Any]], Dict[Any, Any], Any]]: full destination path and
                                                                          data to be written
        """
        # confirm the payload keys are matching accurately with what is expected
        if not {"data", "date"}.issubset(set(payload.keys())):
            raise KeyError("Invalid payload ensure you have date and data as keys")

        if not re.search(r"\d{4}\-\d{2}\-\d{2}", payload["date"]):
            raise ValueError(
                "Invalid date format for partitioning expected: 'YYYY-mm-dd'"
            )
        if not isinstance(payload["data"], list):
            raise TypeError("Invalid data passed: expected List[Dict, Dict]")
        _date = payload["date"].replace("-", "")
        _data = payload["data"]
        write_path = f"{self.folder_path}/{_date}"
        return write_path, _data
